from __future__ import division
import numpy
from mmkfeatures.speech.sigproc import *
from scipy.fftpack import dct

def mfcc(signal,samplerate=16000,winlen=0.025,winstep=0.01,numcep=13,
         nfilt=26,nfft=512,lowfreq=0,highfreq=None,preemph=0.97,ceplifter=22,appendEnergy=True,
         winfunc=lambda x:numpy.ones((x,))):
    """Compute MFCC features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the samplerate of the signal we are working with.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param numcep: the number of cepstrum to return, default 13
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is 512.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :param ceplifter: apply a lifter to final cepstral coefficients. 0 is no lifter. Default is 22.
    :param appendEnergy: if this is true, the zeroth cepstral coefficient is replaced with the log of the total frame energy.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied. You can use numpy window functions here e.g. winfunc=numpy.hamming
    :returns: A numpy array of size (NUMFRAMES by numcep) containing features. Each row holds 1 feature vector.
    """
    feat,energy = fbank(signal,samplerate,winlen,winstep,nfilt,nfft,lowfreq,highfreq,preemph,winfunc)
    feat = numpy.log(feat)
    feat = dct(feat, type=2, axis=1, norm='ortho')[:,:numcep]
    feat = lifter(feat,ceplifter)
    if appendEnergy: feat[:,0] = numpy.log(energy) # replace first cepstral coefficient with log of frame energy
    return feat

def fbank(signal,samplerate=16000,winlen=0.025,winstep=0.01,
          nfilt=26,nfft=512,lowfreq=0,highfreq=None,preemph=0.97,
          winfunc=lambda x:numpy.ones((x,))):
    """Compute Mel-filterbank energy features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the samplerate of the signal we are working with.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is 512.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied. You can use numpy window functions here e.g. winfunc=numpy.hamming
    :returns: 2 values. The first is a numpy array of size (NUMFRAMES by nfilt) containing features. Each row holds 1 feature vector. The
        second return value is the energy in each frame (total energy, unwindowed)
    """
    highfreq= highfreq or samplerate/2
    signal = preemphasis(signal,preemph)
    frames = framesig(signal, winlen*samplerate, winstep*samplerate, winfunc)
    pspec = powspec(frames,nfft)
    energy = numpy.sum(pspec,1) # this stores the total energy in each frame
    energy = numpy.where(energy == 0,numpy.finfo(float).eps,energy) # if energy is zero, we get problems with log

    fb = get_filterbanks(nfilt,nfft,samplerate,lowfreq,highfreq)
    feat = numpy.dot(pspec,fb.T) # compute the filterbank energies
    feat = numpy.where(feat == 0,numpy.finfo(float).eps,feat) # if feat is zero, we get problems with log

    return feat,energy

def logfbank(signal,samplerate=16000,winlen=0.025,winstep=0.01,
          nfilt=26,nfft=512,lowfreq=0,highfreq=None,preemph=0.97):
    """Compute log Mel-filterbank energy features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the samplerate of the signal we are working with.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is 512.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :returns: A numpy array of size (NUMFRAMES by nfilt) containing features. Each row holds 1 feature vector.
    """
    feat,energy = fbank(signal,samplerate,winlen,winstep,nfilt,nfft,lowfreq,highfreq,preemph)
    return numpy.log(feat)

def ssc(signal,samplerate=16000,winlen=0.025,winstep=0.01,
        nfilt=26,nfft=512,lowfreq=0,highfreq=None,preemph=0.97,
        winfunc=lambda x:numpy.ones((x,))):
    """Compute Spectral Subband Centroid features from an audio signal.

    :param signal: the audio signal from which to compute features. Should be an N*1 array
    :param samplerate: the samplerate of the signal we are working with.
    :param winlen: the length of the analysis window in seconds. Default is 0.025s (25 milliseconds)
    :param winstep: the step between successive windows in seconds. Default is 0.01s (10 milliseconds)
    :param nfilt: the number of filters in the filterbank, default 26.
    :param nfft: the FFT size. Default is 512.
    :param lowfreq: lowest band edge of mel filters. In Hz, default is 0.
    :param highfreq: highest band edge of mel filters. In Hz, default is samplerate/2
    :param preemph: apply preemphasis filter with preemph as coefficient. 0 is no filter. Default is 0.97.
    :param winfunc: the analysis window to apply to each frame. By default no window is applied. You can use numpy window functions here e.g. winfunc=numpy.hamming
    :returns: A numpy array of size (NUMFRAMES by nfilt) containing features. Each row holds 1 feature vector.
    """
    highfreq= highfreq or samplerate/2
    signal = preemphasis(signal,preemph)
    frames = framesig(signal, winlen*samplerate, winstep*samplerate, winfunc)
    pspec = powspec(frames,nfft)
    pspec = numpy.where(pspec == 0,numpy.finfo(float).eps,pspec) # if things are all zeros we get problems

    fb = get_filterbanks(nfilt,nfft,samplerate,lowfreq,highfreq)
    feat = numpy.dot(pspec,fb.T) # compute the filterbank energies
    R = numpy.tile(numpy.linspace(1,samplerate/2,numpy.size(pspec,1)),(numpy.size(pspec,0),1))

    return numpy.dot(pspec*R,fb.T) / feat

def hz2mel(hz):
    """Convert a value in Hertz to Mels

    :param hz: a value in Hz. This can also be a numpy array, conversion proceeds element-wise.
    :returns: a value in Mels. If an array was passed in, an identical sized array is returned.
    """
    return 2595 * numpy.log10(1+hz/700.)

def mel2hz(mel):
    """Convert a value in Mels to Hertz

    :param mel: a value in Mels. This can also be a numpy array, conversion proceeds element-wise.
    :returns: a value in Hertz. If an array was passed in, an identical sized array is returned.
    """
    return 700*(10**(mel/2595.0)-1)

def get_filterbanks(nfilt=20,nfft=512,samplerate=16000,lowfreq=0,highfreq=None):
    """Compute a Mel-filterbank. The filters are stored in the rows, the columns correspond
    to fft bins. The filters are returned as an array of size nfilt * (nfft/2 + 1)

    :param nfilt: the number of filters in the filterbank, default 20.
    :param nfft: the FFT size. Default is 512.
    :param samplerate: the samplerate of the signal we are working with. Affects mel spacing.
    :param lowfreq: lowest band edge of mel filters, default 0 Hz
    :param highfreq: highest band edge of mel filters, default samplerate/2
    :returns: A numpy array of size nfilt * (nfft/2 + 1) containing filterbank. Each row holds 1 filter.
    """
    highfreq= highfreq or samplerate/2
    assert highfreq <= samplerate/2, "highfreq is greater than samplerate/2"

    # compute points evenly spaced in mels
    lowmel = hz2mel(lowfreq)
    highmel = hz2mel(highfreq)
    melpoints = numpy.linspace(lowmel,highmel,nfilt+2)
    # our points are in Hz, but we use fft bins, so we have to convert
    #  from Hz to fft bin number
    bin = numpy.floor((nfft+1)*mel2hz(melpoints)/samplerate)

    fbank = numpy.zeros([nfilt,nfft//2+1])
    for j in range(0,nfilt):
        for i in range(int(bin[j]), int(bin[j+1])):
            fbank[j,i] = (i - bin[j]) / (bin[j+1]-bin[j])
        for i in range(int(bin[j+1]), int(bin[j+2])):
            fbank[j,i] = (bin[j+2]-i) / (bin[j+2]-bin[j+1])
    return fbank

def lifter(cepstra, L=22):
    """Apply a cepstral lifter the the matrix of cepstra. This has the effect of increasing the
    magnitude of the high frequency DCT coeffs.

    :param cepstra: the matrix of mel-cepstra, will be numframes * numcep in size.
    :param L: the liftering coefficient to use. Default is 22. L <= 0 disables lifter.
    """
    if L > 0:
        nframes,ncoeff = numpy.shape(cepstra)
        n = numpy.arange(ncoeff)
        lift = 1 + (L/2.)*numpy.sin(numpy.pi*n/L)
        return lift*cepstra
    else:
        # values of L <= 0, do nothing
        return cepstra

def delta(feat, N):
    """Compute delta features from a feature vector sequence.

    :param feat: A numpy array of size (NUMFRAMES by number of features) containing features. Each row holds 1 feature vector.
    :param N: For each frame, calculate delta features based on preceding and following N frames
    :returns: A numpy array of size (NUMFRAMES by number of features) containing delta features. Each row holds 1 delta feature vector.
    """
    if N < 1:
        raise ValueError('N must be an integer >= 1')
    NUMFRAMES = len(feat)
    denominator = 2 * sum([i**2 for i in range(1, N+1)])
    delta_feat = numpy.empty_like(feat)
    padded = numpy.pad(feat, ((N, N), (0, 0)), mode='edge')   # padded version of feat
    for t in range(NUMFRAMES):
        delta_feat[t] = numpy.dot(numpy.arange(-N, N+1), padded[t : t+2*N+1]) / denominator   # [t : t+2*N+1] == [(N+t)-N : (N+t)+N+1]
    return delta_feat



import os
# import traceback
from typing import Dict, Tuple, Union

import cv2
import numpy as np
import torch
from mmkfeatures.video.models.i3d.i3d_src.i3d_net import I3D
from mmkfeatures.video.models.i3d.transforms.transforms import (Clamp, PermuteAndUnsqueeze,
                                              PILToTensor, ResizeImproved,
                                              ScaleTo1_1, TensorCenterCrop,
                                              ToFloat, ToUInt8)
from mmkfeatures.video.models.raft.raft_src.raft import RAFT, InputPadder
from torchvision import transforms
from tqdm import tqdm
from mmkfeatures.video.utils.utils import (action_on_extraction, form_list_from_user_input,
                         reencode_video_with_diff_fps,
                         show_predictions_on_dataset)

current_path = os.path.dirname(os.path.realpath(__file__)).replace("\\i3d","")

PWC_MODEL_PATH = current_path+ '/pwc/checkpoints/pwc_net_sintel.pt'
RAFT_MODEL_PATH =  current_path+ '/models/raft/checkpoints/raft-sintel.pth'
I3D_RGB_PATH =  current_path+ '/i3d/checkpoints/i3d_rgb.pt'
I3D_FLOW_PATH =  current_path+ '/i3d/checkpoints/i3d_flow.pt'
PRE_CENTRAL_CROP_MIN_SIDE_SIZE = 256
CENTRAL_CROP_MIN_SIDE_SIZE = 224
DEFAULT_I3D_STEP_SIZE = 64
DEFAULT_I3D_STACK_SIZE = 64
I3D_CLASSES_NUM = 400

class ExtractI3D(torch.nn.Module):

    def __init__(self, args):
        super(ExtractI3D, self).__init__()
        self.feature_type = args.feature_type
        if args.streams is None:
            self.streams = ['rgb', 'flow']
        else:
            self.streams = [args.streams]
        self.path_list = form_list_from_user_input(args)
        self.flow_type = args.flow_type
        self.flow_model_paths = {'pwc': PWC_MODEL_PATH, 'raft': RAFT_MODEL_PATH}
        self.i3d_weights_paths = {'rgb': I3D_RGB_PATH, 'flow': I3D_FLOW_PATH}
        self.min_side_size = PRE_CENTRAL_CROP_MIN_SIDE_SIZE
        self.central_crop_size = CENTRAL_CROP_MIN_SIDE_SIZE
        self.extraction_fps = args.extraction_fps
        self.step_size = args.step_size
        self.stack_size = args.stack_size
        if self.step_size is None:
            self.step_size = DEFAULT_I3D_STEP_SIZE
        if self.stack_size is None:
            self.stack_size = DEFAULT_I3D_STACK_SIZE
        self.resize_transforms = transforms.Compose([
            transforms.ToPILImage(),
            ResizeImproved(self.min_side_size),
            PILToTensor(),
            ToFloat(),
        ])
        self.i3d_transforms = {
            'rgb': transforms.Compose([
                TensorCenterCrop(self.central_crop_size),
                ScaleTo1_1(),
                PermuteAndUnsqueeze()
            ]),
            'flow': transforms.Compose([
                TensorCenterCrop(self.central_crop_size),
                Clamp(-20, 20),
                ToUInt8(),
                ScaleTo1_1(),
                PermuteAndUnsqueeze()
            ])
        }
        self.show_pred = args.show_pred
        self.i3d_classes_num = I3D_CLASSES_NUM
        self.keep_tmp_files = args.keep_tmp_files
        self.on_extraction = args.on_extraction
        self.tmp_path = os.path.join(args.tmp_path, self.feature_type)
        self.output_path = os.path.join(args.output_path, self.feature_type)
        self.progress = tqdm(total=len(self.path_list))

    def forward(self, indices: torch.LongTensor):
        '''
        Arguments:
            indices {torch.LongTensor} -- indices to self.path_list
        '''
        device = indices.device
        flow_xtr_model, models = self.load_model(device)

        for idx in indices:
            # when error occurs might fail silently when run from torch data parallel
            try:
                feats_dict = self.extract(device, flow_xtr_model, models, self.path_list[idx])
                action_on_extraction(feats_dict, self.path_list[idx], self.output_path, self.on_extraction)
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except Exception as e:
                # traceback.print_exc()  # for the whole traceback
                print(e)
                print(f'Extraction failed at: {self.path_list[idx]}. Continuing extraction')

            # update tqdm progress bar
            self.progress.update()

    def extract(self,
                device: torch.device,
                flow_xtr_model: torch.nn.Module,
                models: Dict[str, torch.nn.Module],
                video_path: Union[str, None] = None
                ) -> Dict[str, Union[torch.nn.Module, str]]:
        '''The extraction call. Made to clean the forward call a bit.

        Arguments:
            device {torch.device}
            flow_xtr_model {torch.nn.Module}
            models {Dict[str, torch.nn.Module]}

        Keyword Arguments:
            video_path {Union[str, None]} -- if you would like to use import it and use it as
                                             "path -> model"-fashion (default: {None})

        Returns:
            Dict[str, Union[torch.nn.Module, str]] -- dict with i3d features and their type
        '''
        def _run_on_a_stack(feats_dict, rgb_stack, models, device, stack_counter, padder=None):
            rgb_stack = torch.cat(rgb_stack).to(device)

            for stream in self.streams:
                with torch.no_grad():
                    # if i3d stream is flow, we first need to calculate optical flow, otherwise, we use rgb
                    # `end_idx-1` and `start_idx+1` because flow is calculated between f and f+1 frames
                    # we also use `end_idx-1` for stream == 'rgb' case: just to make sure the feature length
                    # is same regardless of whether only rgb is used or flow
                    if stream == 'flow':
                        if self.flow_type == 'raft':
                            stream_slice = flow_xtr_model(padder.pad(rgb_stack)[:-1], padder.pad(rgb_stack)[1:])
                        elif self.flow_type == 'pwc':
                            stream_slice = flow_xtr_model(rgb_stack[:-1], rgb_stack[1:])
                        else:
                            raise NotImplementedError
                    elif stream == 'rgb':
                        stream_slice = rgb_stack[:-1]
                    else:
                        raise NotImplementedError
                    # apply transforms depending on the stream (flow or rgb)
                    stream_slice = self.i3d_transforms[stream](stream_slice)
                    # extract features for a stream
                    feats = models[stream](stream_slice, features=True)  # (B, 1024)
                    # add features to the output dict
                    feats_dict[stream].extend(feats.tolist())
                    # show predictions on a daataset
                    if self.show_pred:
                        softmaxes, logits = models[stream](stream_slice, features=False)  # (B, classes=400)
                        print(f'{video_path} @ stack {stack_counter} ({stream} stream)')
                        show_predictions_on_dataset(logits, 'kinetics')

        # take the video, change fps and save to the tmp folder
        if self.extraction_fps is not None:
            video_path = reencode_video_with_diff_fps(video_path, self.tmp_path, self.extraction_fps)

        cap = cv2.VideoCapture(video_path)
        fps = cap.get(cv2.CAP_PROP_FPS)
        # timestamp when the last frame in the stack begins (when the old frame of the last pair ends)
        timestamps_ms = []
        stack = []
        feats_dict = {stream: [] for stream in self.streams}

        # sometimes when the target fps is 1 or 2, the first frame of the reencoded video is missing
        # and cap.read returns None but the rest of the frames are ok. timestep is 0.0 for the 2nd frame in
        # this case
        first_frame = True
        padder = None
        stack_counter = 0
        while cap.isOpened():
            frame_exists, rgb = cap.read()

            if first_frame:
                first_frame = False
                if frame_exists is False:
                    continue

            if frame_exists:
                # preprocess the image
                rgb = cv2.cvtColor(rgb, cv2.COLOR_BGR2RGB)
                rgb = self.resize_transforms(rgb)
                rgb = rgb.unsqueeze(0)

                if self.flow_type == 'raft' and padder is None:
                    padder = InputPadder(rgb.shape)

                stack.append(rgb)

                # - 1 is used because we need B+1 frames to calculate B frames
                if len(stack) - 1 == self.stack_size:
                    _run_on_a_stack(feats_dict, stack, models, device, stack_counter, padder)
                    # leaving the elements if step_size < stack_size so they will not be loaded again
                    # if step_size == stack_size one element is left because the flow between the last element
                    # in the prev list and the first element in the current list
                    stack = stack[self.step_size:]
                    stack_counter += 1
                    timestamps_ms.append(cap.get(cv2.CAP_PROP_POS_MSEC))
            else:
                # we don't run inference if the stack is not full (applicable for i3d)
                cap.release()
                break

        # removes the video with different fps if it was created to preserve disk space
        if (self.extraction_fps is not None) and (not self.keep_tmp_files):
            os.remove(video_path)

        # transforms list of features into a np array
        feats_dict = {stream: np.array(feats) for stream, feats in feats_dict.items()}
        # also include the timestamps and fps
        feats_dict['fps'] = np.array(fps)
        feats_dict['timestamps_ms'] = np.array(timestamps_ms)

        return feats_dict

    def load_model(self, device: torch.device) -> Tuple[torch.nn.Module, Dict[str, torch.nn.Module]]:
        '''Defines the models, loads checkpoints, sends them to the device.

        Args:
            device (torch.device): The device

        Raises:
            NotImplementedError: if flow type is not implemented.

        Returns:
            Tuple[torch.nn.Module, Dict[str, torch.nn.Module]]: flow extraction module and the model.
        '''
        # Flow extraction module
        if self.flow_type == 'pwc':
            from mmkfeatures.video.models.pwc.pwc_src.pwc_net import PWCNet
            flow_xtr_model = PWCNet()
        elif self.flow_type == 'raft':
            flow_xtr_model = RAFT()
            flow_xtr_model = torch.nn.DataParallel(flow_xtr_model, device_ids=[device])
        else:
            raise NotImplementedError

        flow_xtr_model.load_state_dict(torch.load(self.flow_model_paths[self.flow_type], map_location='cpu'))
        flow_xtr_model = flow_xtr_model.to(device)
        flow_xtr_model.eval()

        # Feature extraction models (rgb and flow streams)
        i3d_stream_models = {}
        for stream in self.streams:
            i3d_stream_model = I3D(num_classes=self.i3d_classes_num, modality=stream)
            i3d_stream_model.load_state_dict(torch.load(self.i3d_weights_paths[stream], map_location='cpu'))
            i3d_stream_model = i3d_stream_model.to(device)
            i3d_stream_model.eval()
            i3d_stream_models[stream] = i3d_stream_model

        return flow_xtr_model, i3d_stream_models



from sklearn.feature_extraction.text import CountVectorizer,TfidfVectorizer
from spacy.lang.en.stop_words import STOP_WORDS
import os
from gensim.scripts.glove2word2vec import glove2word2vec
from mmkfeatures.text.featurization import Bow
from gensim.models import KeyedVectors
from gensim.models import Word2Vec
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import word_tokenize
class TextFeaturesWrapper:

    def __init__(self):
        pass

    def get_text_features_using_tfidf(self,corpus, text):
        # Create CountVectorizer, which create bag-of-words model.
        # stop_words : Specify language to remove stopwords.
        vectorizer = TfidfVectorizer(stop_words='english')

        # Learn vocabulary in sentences.
        vectorizer.fit(corpus)

        # Get dictionary.
        vectorizer.get_feature_names()

        # Transform each sentences in vector space.
        if type(text) == str:
            input_text = [text]
        else:
            input_text = text
        vector = vectorizer.transform(input_text)
        vector_spaces = vector.toarray()
        if type(text) == str:
            return vector_spaces[0]
        else:
            return vector_spaces

    def get_text_features_using_bow(self,corpus, text):
        # Create CountVectorizer, which create bag-of-words model.
        # stop_words : Specify language to remove stopwords.
        vectorizer = CountVectorizer(stop_words='english')

        # Learn vocabulary in sentences.
        vectorizer.fit(corpus)

        # Get dictionary.
        vectorizer.get_feature_names()

        # Transform each sentences in vector space.
        if type(text) == str:
            input_text = [text]
        else:
            input_text = text

        vector = vectorizer.transform(input_text)
        vector_spaces = vector.toarray()
        if type(text) == str:
            return vector_spaces[0]
        else:
            return vector_spaces

    def tokenize(self,sentence):
        words=[token.lower() for token in word_tokenize(sentence) if token not in STOP_WORDS]
        wnl = WordNetLemmatizer()
        tokens=[wnl.lemmatize(wrd) for wrd in words]
        return tokens


    def get_text_features_using_bow2(self,corpus, text):
        # sentences = ['this is a list of sentences', 'second sentence in list of sentences', 'a word for complexity']
        bow = Bow()
        bow.fit(corpus)
        bow.transform(corpus)
        vs = bow.transform(text)
        return vs

    def get_text_features_using_wordvector(self,corpus, text):

        new_corpus = []
        for c in corpus:
            new_corpus.append(word_tokenize(c))

        model = Word2Vec(sentences=new_corpus, vector_size=100, window=5, min_count=1, workers=4)
        # model.save("word2vec.model")
        # model = Word2Vec.load("word2vec.model")

        # model.train(new_corpus, total_examples=1, epochs=1)
        # vector = model.wv['computer']  # get numpy vector of a word
        # sims = model.wv.most_similar('computer', topn=10)  # get other similar words
        sen_tokens = self.tokenize(text)
        print("tokens: ", sen_tokens)
        vs = []
        for w in sen_tokens:
            vs.append(model.wv[w])
        return vs

    def get_text_features_using_GloVe(self,text, glove_input_file, word2vec_output_file):
        if not os.path.exists(word2vec_output_file):
            print("Converting glove file to wordvec file... ")
            (count, dimensions) = glove2word2vec(glove_input_file, word2vec_output_file)
        # load model
        glove_model = KeyedVectors.load_word2vec_format(word2vec_output_file, binary=False)

        sen_tokens = self.tokenize(text)
        print("tokens: ", sen_tokens)
        vs = []
        for w in sen_tokens:
            vs.append(glove_model[w])
        return vs
        # obtain the vector of a single word
        # cat_vec = glove_model['cat']
        # print(cat_vec)
        # get similar words
        # print(glove_model.most_similar('frog'))
        # return cat_vec

if __name__ == "__main__":

    text_features_wrapper=TextFeaturesWrapper()

    corpus = [
        "This is a sample sentence",
        "I am interested in politics",
        "You are a very good software engineer, engineer.",
    ]

    sentence= "It is a sample sentence"

    # vs=get_text_features_using_tfidf(corpus,[sentence,sentence])
    # vs = get_text_features_using_wordembedding(corpus, sentence)
    # vs=get_text_features_using_GloVe(sentence,"../../data/glove.6B/glove.6B.50d.txt","../../data/glove.6B/glove_6B.50d.wordvec.txt")

    vs=text_features_wrapper.get_text_features_using_wordvector(corpus,sentence)
    print(vs)



featuresetMetadataTemplate=	[
			"root name",#name of the featureset
			"computational sequence description",#name of the featureset
			"dimension names"
			"computational sequence version",#the version of featureset
			"alignment compatible",#name of the featureset
			"dataset name",#featureset belongs to which dataset
			"dataset version",#the version of dataset
			"creator",#the author of the featureset
			"contact",#the contact of the featureset
			"featureset bib citation",#citation for the paper related to this featureset
			"dataset bib citation"#citation for the dataset
			]






import os
from mmkfeatures.image.deep_extractor import DeepExtractor
from mmkfeatures.image.lbp_extractor import LBPExtractor
from mmkfeatures.image.bow_extractor import BoWExtractor

# model_name options: ['vgg19', 'inception_v3', 'inception_resnet_v2']

class ImageFeaturesWrapper:
    def __init__(self):
        pass

    def get_deep_feature_extractor(self,src,dest=None,model_name="vgg19"):
        return DeepExtractor(base_route=src, model_name=model_name, size=75)

    def extract_deep_features(self,src, extractor=None,dest=None,model_name="vgg19"):
        if extractor==None:
            extractor = DeepExtractor(base_route=src, model_name=model_name, size=75)

        features=[]
        if dest==None:
            if os.path.isfile(src):
                features = [extractor.extract(src)]
            elif os.path.isdir(src):
                features = []
                list = os.listdir(src)  # 列出文件夹下所有的目录与文件
                for i in range(0, len(list)):
                    path = os.path.join(src, list[i])
                    if os.path.isfile(path):
                        features.append(extractor.extract(path))
            else:
                raise Exception("No file path is recognized!")
            return features
        else:
            if not os.path.isdir(src):
                raise Exception("src should be a directory when dest has value. ")

            extractor.extract_and_save(dest)
            return features


    def extract_lbp_features(self,src,dest=None,size=64,points=3,radius=1,x=2,y=4):

        extractor = LBPExtractor(base_route=src, size=size, points=points, radius=radius, grid_x=x, grid_y=y)

        features = []
        if dest==None:
            if os.path.isfile(src):
                features = [extractor.extract(src)]
            elif os.path.isdir(src):
                features = []
                list = os.listdir(src)  # 列出文件夹下所有的目录与文件
                for i in range(0, len(list)):
                    path = os.path.join(src, list[i])
                    if os.path.isfile(path):
                        features.append(extractor.extract(path))
            else:
                raise Exception("No file path is recognized!")
            return features
        else:
            if not os.path.isdir(src):
                raise Exception("src should be a directory when dest has value. ")

            extractor.extract_and_save(dest)
            return features

    def extract_bow_features(self,src_dir,src,dest,k=4, method="kaze", size=64, cluster_mode='manual'):

        extractor = BoWExtractor(base_route=src_dir, k=k, method=method, size=size,
                                 cluster_mode=cluster_mode)
        extractor.setup()
        extractor.fit()

        features = []
        if dest==None:
            if os.path.isfile(src):
                features = [extractor.extract(src)]
            elif os.path.isdir(src):
                features = []
                list = os.listdir(src)  # 列出文件夹下所有的目录与文件
                for i in range(0, len(list)):
                    path = os.path.join(src, list[i])
                    if os.path.isfile(path):
                        features.append(extractor.extract(path))
            else:
                raise Exception("No file path is recognized!")
            return features
        else:
            if not os.path.isdir(src):
                raise Exception("src should be a directory when dest has value. ")
            extractor.extract_and_save(dest)
            return features

if __name__ == '__main__':
    img_feat_wrapper=ImageFeaturesWrapper()
    # 1. extract a single feature from a file
    src_tiny_imagenet="../dataset/micro-imagenet/train"
    src="../dataset/micro-imagenet/train/n03014705/images/n03014705_0.JPEG"
    dest=None
    # features=img_feat_wrapper.extract_lbp_features(src,dest)
    # features=img_feat_wrapper.extract_bow_features(src,dest)
    features=img_feat_wrapper.extract_bow_features(src_tiny_imagenet,src,dest)
    print(features[0].shape)
    print(features)

    # 2. extract features from a folder and save the features to a file
    src="../dataset/micro-imagenet/train"
    dest="../dataset/output/feature.csv"
    # features=img_feat_wrapper.extract_lbp_features(src,dest)
    features = img_feat_wrapper.extract_bow_features(src_tiny_imagenet,src, dest)




"""
特征提取函数集合
"""
import numpy
from scipy.fftpack import fft
from scipy.fftpack.realtransforms import dct
from librosa.feature import mfcc
EPS = numpy.finfo(float).eps

def get_feature(fs, signal,n_mfcc):
    """
    简易的提取特征函数
    :param fs: 采样率
    :param signal: 信号
    :return feature:mfcc特征
    """
    feature = mfcc(signal, fs, S=None, n_mfcc=n_mfcc).T
    return feature

def mt_feature_extraction(signal, fs, mt_win, mt_step, st_win):
    """
    提取 mid-term feature和short-term feature
    :param signal: 信号
    :param fs: 采样率
    :param mtWin: mid-term窗口大小
    :param mtStep: mid-term步长
    :param stWin: short-term窗口大小
    :param stStep: short-term步长
    :return: mid-term feature, short-term feature
    """
    st_step = st_win
    mt_win_ratio = int(round(mt_win / st_step))
    mt_step_ratio = int(round(mt_step / st_step))

    st_features = st_feature_extraction(signal, fs, st_win, st_step)
    num_of_features = len(st_features)
    num_of_statistics = 2

    mt_features = []
    for i in range(num_of_statistics * num_of_features):
        mt_features.append([])

    for i in range(num_of_features):
        cur_pos = 0
        len_feat = len(st_features[i])
        while cur_pos < len_feat:
            pos_1 = cur_pos
            pos_2 = cur_pos + mt_win_ratio
            if pos_2 > len_feat:
                pos_2 = len_feat
            cur_st_features = st_features[i][pos_1:pos_2]

            mt_features[i].append(numpy.mean(cur_st_features))
            mt_features[i+num_of_features].append(numpy.std(cur_st_features))
            #mtFeatures[i+2*numOfFeatures].append(numpy.std(cur_st_features) /
            # (numpy.mean(cur_st_features)+EPS))
            cur_pos += mt_step_ratio

    return numpy.array(mt_features), st_features


def st_feature_extraction(signal, fs, win, step):
    """
    提取 short-term feature
    :param signal: 信号
    :param fs: 采样率
    :param win: short-term窗口大小
    :param step: short-term步长
    :return: short-term feature
    """
    win = int(win)
    step = int(step)
    signal = numpy.double(signal)
    signal = signal / (2.0 ** 15)
    dc = signal.mean()
    max_value = (numpy.abs(signal)).max()
    signal = (signal - dc) / (max_value + EPS)

    len_feat = len(signal)
    cur_pos = 0
    count_frames = 0
    n_fft = win / 2

    [fbank, _] = mfcc_init_filter_banks(fs, n_fft)
    n_chroma, n_freqs_per_chroma = st_chroma_features_init(n_fft, fs)

    num_of_time_spectral_features = 8
    num_of_harmonic_features = 0
    nceps = 13
    num_of_chroma_features = 13
    total_num_of_features = num_of_time_spectral_features + nceps + num_of_harmonic_features + \
                         num_of_chroma_features
#    totalNumOfFeatures = num_of_time_spectral_features + nceps + numOfHarmonicFeatures

    st_features = []
    xprev = []
    while cur_pos + win - 1 < len_feat:
        count_frames += 1
        pos_signal = signal[cur_pos:cur_pos+win]
        cur_pos = cur_pos + step
        cur_pos_signal = abs(fft(pos_signal))
        cur_pos_signal = cur_pos_signal[0:int(n_fft)]
        cur_pos_signal = cur_pos_signal / len(cur_pos_signal)
        if count_frames == 1:
            xprev = cur_pos_signal.copy()
        cur_vf = numpy.zeros((total_num_of_features, 1))
        cur_vf[0] = st_zcr(pos_signal)
        cur_vf[1] = st_energy(pos_signal)
        cur_vf[2] = st_energy_entropy(pos_signal)
        [cur_vf[3], cur_vf[4]] = st_spectral_centroid_and_spread(cur_pos_signal, fs)
        cur_vf[5] = st_spectral_entropy(cur_pos_signal)
        cur_vf[6] = st_spectral_flux(cur_pos_signal, xprev)
        cur_vf[7] = st_pectral_roll_off(cur_pos_signal, 0.90, fs)
        cur_vf[num_of_time_spectral_features:num_of_time_spectral_features+nceps, 0] = \
            st_mfcc(cur_pos_signal, fbank, nceps).copy()
        _, chroma_f = st_chroma_features(cur_pos_signal, n_chroma, n_freqs_per_chroma)
        cur_vf[num_of_time_spectral_features + nceps:
               num_of_time_spectral_features + nceps + num_of_chroma_features - 1] = chroma_f
        cur_vf[num_of_time_spectral_features + nceps + num_of_chroma_features - 1] = chroma_f.std()
        st_features.append(cur_vf)
        xprev = cur_pos_signal.copy()

    st_features = numpy.concatenate(st_features, 1)
    return st_features


def st_zcr(frame):
    """
    过零率
    """
    count = len(frame)
    count_z = numpy.sum(numpy.abs(numpy.diff(numpy.sign(frame)))) / 2
    return numpy.float64(count_z) / numpy.float64(count-1.0)


def st_energy(frame):
    """
    短时能量
    """
    return numpy.sum(frame ** 2) / numpy.float64(len(frame))


def st_energy_entropy(frame, num_of_short_blocks=10):
    """
    短时能量熵
    """
    eol = numpy.sum(frame ** 2)
    len_frame = len(frame)
    sub_win_length = int(numpy.floor(len_frame / num_of_short_blocks))
    if len_frame != sub_win_length * num_of_short_blocks:
        frame = frame[0:sub_win_length * num_of_short_blocks]
    sub_windows = frame.reshape(sub_win_length, num_of_short_blocks, order='F').copy()
    sum_windows = numpy.sum(sub_windows ** 2, axis=0) / (eol + EPS)
    neg_entropy = numpy.sum(sum_windows * numpy.log2(sum_windows + EPS))
    return neg_entropy if neg_entropy else -1


def st_spectral_centroid_and_spread(cur_pos_signal, fs):
    """
    短时谱速度
    """
    ind = (numpy.arange(1, len(cur_pos_signal) + 1)) * (fs/(2.0 * len(cur_pos_signal)))
    cur_signal = cur_pos_signal.copy()
    cur_signal = cur_signal / cur_signal.max()
    num = numpy.sum(ind * cur_signal)
    den = numpy.sum(cur_signal) + EPS
    cen = (num / den)
    sum_windows = numpy.sqrt(numpy.sum(((ind - cen) ** 2) * cur_signal) / den)
    cen = cen / (fs / 2.0)
    sum_windows = sum_windows / (fs / 2.0)
    return cen, sum_windows


def st_spectral_entropy(cur_pos_signal, num_of_short_blocks=10):
    """
    短时谱熵
    """
    len_frame = len(cur_pos_signal)
    eol = numpy.sum(cur_pos_signal ** 2)
    sub_win_length = int(numpy.floor(len_frame / num_of_short_blocks))
    if len_frame != sub_win_length * num_of_short_blocks:
        cur_pos_signal = cur_pos_signal[0:sub_win_length * num_of_short_blocks]
    sub_windows = cur_pos_signal.reshape(sub_win_length, num_of_short_blocks,
                                         order='F').copy()
    sum_windows = numpy.sum(sub_windows ** 2, axis=0) / (eol + EPS)
    ne_entropy = numpy.sum(sum_windows*numpy.log2(sum_windows + EPS))
    return -ne_entropy if ne_entropy else -1


def st_spectral_flux(cur_pos_signal, xprev):
    """
    短时谱通量
    """
    sum_x = numpy.sum(cur_pos_signal + EPS)
    sum_prev_x = numpy.sum(xprev + EPS)
    st_flux = numpy.sum((cur_pos_signal / sum_x - xprev/sum_prev_x) ** 2)
    return st_flux


def st_pectral_roll_off(cur_pos_signal, cen, fs):
    """
    短时谱滚动量
    """
    total_energy = numpy.sum(cur_pos_signal ** 2)
    fft_length = len(cur_pos_signal)
    thres = cen*total_energy
    cum_sum = numpy.cumsum(cur_pos_signal ** 2) + EPS
    [no_neg, ] = numpy.nonzero(cum_sum > thres)
    if no_neg[0]:
        m_c = numpy.float64(no_neg[0]) / (float(fft_length))
    else:
        m_c = 0.0
    return m_c


def mfcc_init_filter_banks(fs, nfft):
    """
    mfcc初始滤波宽度
    """
    lowfreq = 133.33
    linsc = 200 / 3.
    logsc = 1.0711703
    num_lin_filt_total = 13
    num_log_filt = 27

    n_filt_total = num_lin_filt_total + num_log_filt
    freqs = numpy.zeros(n_filt_total + 2)
    freqs[:num_lin_filt_total] = lowfreq + numpy.arange(num_lin_filt_total) * linsc
    freqs[num_lin_filt_total:] = freqs[num_lin_filt_total - 1] * \
                                 logsc ** numpy.arange(1, num_log_filt + 3)
    heights = 2. / (freqs[2:] - freqs[0:-2])
    fbank = numpy.zeros((int(n_filt_total), int(nfft)))
    nfreqs = numpy.arange(nfft) / (1. * nfft) * fs

    for i in range(n_filt_total):
        low_tr_freq = freqs[i]
        cen_tr_freq = freqs[i + 1]
        high_tr_freq = freqs[i + 2]

        lid = numpy.arange(numpy.floor(low_tr_freq * nfft / fs) + 1,
                           numpy.floor(cen_tr_freq * nfft / fs) + 1, dtype=numpy.int)
        lslope = heights[i] / (cen_tr_freq - low_tr_freq)
        rid = numpy.arange(numpy.floor(cen_tr_freq * nfft / fs) + 1,
                           numpy.floor(high_tr_freq * nfft / fs) + 1, dtype=numpy.int)
        rslope = heights[i] / (high_tr_freq - cen_tr_freq)
        fbank[i][lid] = lslope * (nfreqs[lid] - low_tr_freq)
        fbank[i][rid] = rslope * (high_tr_freq - nfreqs[rid])

    return fbank, freqs


def st_mfcc(cur_pos_signal, fbank, nceps):
    """
    短时mfcc
    """
    mspec = numpy.log10(numpy.dot(cur_pos_signal, fbank.T) + EPS)
    ceps = dct(mspec, type=2, norm='ortho', axis=-1)[:nceps]
    return ceps


def st_chroma_features_init(nfft, fs):
    """
    色度初始化
    """
    freqs = numpy.array([((st_flux + 1) * fs) / (2 * int(nfft)) for st_flux in range(int(nfft))])
    c_p = 27.50
    n_chroma = numpy.round(12.0 * numpy.log2(freqs / c_p)).astype(int)
    n_freqs_per_chroma = numpy.zeros((n_chroma.shape[0],))
    u_chroma = numpy.unique(n_chroma)
    for u_ch in u_chroma:
        idx = numpy.nonzero(n_chroma == u_ch)
        n_freqs_per_chroma[idx] = idx[0].shape
    return n_chroma, n_freqs_per_chroma


def st_chroma_features(cur_pos_signal, n_chroma, n_freqs_per_chroma):
    """
    短时色度
    """
    chroma_names = ['A', 'A#', 'B', 'cen', 'cen#', 'D', 'D#', 'E', 'st_flux', 'st_flux#', 'G', 'G#']
    spec = cur_pos_signal ** 2
    if n_chroma.max() < n_chroma.shape[0]:
        cen = numpy.zeros((n_chroma.shape[0],))
        cen[n_chroma] = spec
        cen /= n_freqs_per_chroma[n_chroma]
    else:
        no_0_pos = numpy.nonzero(n_chroma > n_chroma.shape[0])[0][0]
        cen = numpy.zeros((n_chroma.shape[0],))
        cen[n_chroma[0:no_0_pos - 1]] = spec
        cen /= n_freqs_per_chroma
    new_d = int(numpy.ceil(cen.shape[0] / 12.0) * 12)
    cur_two = numpy.zeros((new_d,))
    cur_two[0:cen.shape[0]] = cen
    cur_two = cur_two.reshape(int(cur_two.shape[0] / 12), 12)
    final_c = numpy.matrix(numpy.sum(cur_two, axis=0)).T
    final_c /= spec.sum()
    return chroma_names, final_c



import h5py

file_name="../full_examples/cmumosi/" \
          "http___immortal.multicomp.cs.cmu.edu_CMU-MOSI_language_CMU_MOSI_TimestampedPhones.csd"

file_name="../../../examples/compseq_1.csd"

import nexusformat.nexus as nx
f = nx.nxload(file_name)
print(f.tree)

print()
with h5py.File(file_name, "r") as f:
    for key in f.keys():
        print(key)  # Names of the groups in HDF5 file.

    # Get the HDF5 group
    group = f[key]
    # Checkout what keys are inside that group.
    for key in group.keys():
        print(key)

    print(list(group["metadata"]))
    meta_keys=list(group["metadata"])
    for mk in meta_keys:
        print("meta key: ",mk)
        print("-> ",group["metadata"][mk][0])

    print()
    content_ids=list(group["data"])
    print(content_ids)
    for k in content_ids: # each video
        print("[Content ID] -> ",k)
        part=group["data"][k]
        # print(list(part))
        features=part['features']
        intervals=part['intervals']
        # print(features)
        # print(intervals)
        print("[features] ->")
        for f in features[:5]:
            print("--> ",f)
        print("number of features: ",len(features))
        print("[intervals] ->")
        for inter in intervals[:5]:
            print("--> ",inter)
        print("number of intervals: ",len(intervals))
        print()












import h5py
import hashlib
import validators
import json
import sys
import os
import time
import uuid
# log is the same as standard_sdks log
from mmkfeatures.fusion import log
from mmkfeatures.fusion.metadataconfigs import *
from mmkfeatures.fusion.integrity_check import *

from mmkfeatures.fusion.blank import *
from mmkfeatures.fusion.file_ops import *
from mmkfeatures.fusion.download_ops import *

# computational sequence class
# main attributes:
#       main_file: where the location of the heirarchical data binary file is
#       resource: where the heirarchical data comes from (i.e. a URL)
#       h5handle: handle to the file
#       data: the data in a dictionary format
#       metadata: the metadata in a dictionary format
#
# main function:
#       completeAllMissingMetadata: a helper to complete all the missing metadata
#       deploy: deploy the computational sequence - you can also notify multicomp about your computational sequence

class computational_sequence():

    def __init__(self, resource, destination=None, validate=True):
        # initializing the featureset
        self.__initialize(resource, destination, validate)

        # BACKWARD: backward compatibility
        self.rootName = self.root_name

    def _compare_entries(self, entry1, entry2):
        return entry1.split('[')[0] == entry2.split('[')[0]

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = {}
        self.data[key]["intervals"] = value["intervals"]
        self.data[key]["features"] = value["features"]

    def keys(self):
        return self.data.keys()

    def _remove_id(self, entry_id, purge=False):
        if purge == False:
            if entry_id in list(self.data.keys()):
                del self.data[entry_id]
        elif purge == True:
            keys_to_del = [key for key in list(self.keys()) if key[:len(entry_id)] == entry_id]
            for key_to_del in keys_to_del:
                del self.data[key_to_del]
        else:
            log.error("Purge received wrong argument type. Exiting ...!", error=True)

    def __initialize_from_csd(self, resource, destination, validate):
        if validators.url(resource):
            # user would like to store to the current directory
            if destination is None or destination == '':
                destination = os.path.join('./', resource.split('/')[-1])
            # user has chosen a different directory
            elif '.csd' not in destination:
                destination = os.path.join(destination, resource.split('/')[-1])
            # .csd cannot be in the destination as we don't allow name changes when downloading
            else:
                log.error(
                    "Destination needs to be a folder where the downloaded computational sequence is stored. Exiting ...!",
                    error=True)

            read_URL(resource, destination)
            self.resource = resource
            self.main_file = destination
        else:
            self.main_file = resource

        print("self.mainfile=",self.main_file)

        h5handle, data,_,_, metadata = read_CSD(self.main_file)
        if type(metadata) is not dict:
            log.error("Metadata not in correct format for %s. Exiting ...!" % destination, error=True)

        self.data = data
        self.metadata = metadata
        self.root_name = self.metadata["root name"]
        self.h5handle = h5handle

        if validate:
            self.__check_format()
        else:
            log.warning("Validation of the computational sequence skipped by user request.")

    def __initialize_blank(self, root_name):
        self.main_file = None
        self.h5handle = None
        self.root_name = root_name
        self.data = {}
        self.metadata = {}
        self.metadata["root name"] = root_name
        log.success("Initialized empty <%s> computational sequence." % self.metadata["root name"])

    # TODO: try and excepts to be added to this code
    def __initialize(self, resource, destination, validate):
        # computational sequence is already initialized
        if hasattr(self, 'h5handle'): log.error(
            "<%s> computational sequence already initialized ..." % self.metadata["root name"], error=True)

        # initializing blank - main_file is where to initialize the data and resource is None since the data comes from nowhere
        if '.csd' not in resource:
            self.__initialize_blank(resource)
        # reading from url or folder - main_file is where the data should go and resource is the url
        else:
            self.__initialize_from_csd(resource, destination, validate)

    # checking if the data and metadata are in correct format
    # stops the program if the integrity is not ok
    def __check_format(self, error=True):
        if not hasattr(self, 'metadata') or not hasattr(self, 'data'):
            log.error("computational sequence is blank (data or metadata is missing)")
        log.status("Checking the integrity of the <%s> computational sequence ..." % self.metadata["root name"])
        # TODO: hash check not implemented yet
        datavalid = validate_data_format(self.data, self.metadata["root name"], verbose=False)
        metadatavalid = validate_metadata_format(self.metadata, self.metadata["root name"], verbose=False)
        if datavalid and metadatavalid:
            log.success("<%s> computational sequence is valid!" % self.metadata["root name"])

    # set the metadata for all the missing information in the metadata. If the key is not set then the assumption is that metadata is not available. Note that if the metadata is set to a dummy variable like None then it is still considered set.
    def complete_all_missing_metadata(self):

        missings = [x for (x, y) in zip(featuresetMetadataTemplate,
                                        [metadata in self.metadata.keys() for metadata in featuresetMetadataTemplate])
                    if y is False]
        # python2 vs python 3
        # TODO: Add read from file
        root_name_ext = ''
        if hasattr(self, "root_name"):
            root_name_ext = " for <%s> computational sequence" % self.root_name
        for missing in missings:
            self.metadata[missing] = log.status("Please input %s%s: " % (missing, root_name_ext), require_input=True)

    # BACKWARD: _ for backward compability
    def set_data(self, data, _=None, __=None):
        validate_data_format(data, self.root_name, verbose=True)
        self.data = data

    # BACKWARD: _ for backward compability
    def set_metadata(self, metadata, _=None, __=None):
        validate_metadata_format(metadata, self.root_name, verbose=False)
        self.metadata = metadata

    # writing the file to the output
    def deploy(self, destination, compression="gzip", compression_opts=9, full_chunk_shape=True,my_metadata=None):
        if my_metadata == None:
            self.complete_all_missing_metadata()
        else:
            self.metadata = my_metadata

        self.__check_format()
        log.status("Deploying the <%s> computational sequence to %s" % (destination, self.metadata['root name']))
        # generating the unique identifiers
        self.metadata['uuid'] = str(uuid.uuid4())
        # TODO: add SHA256 check + midification should not be possible without private key
        self.metadata['md5'] = None
        log.status("Your unique identifier for <%s> computational sequence is %s" % (
        self.metadata["root name"], self.metadata['uuid']))
        write_CSD(self.data,{}, {}, self.metadata, self.metadata["root name"], destination, compression=compression,
                  compression_opts=compression_opts, full_chunk_shape=full_chunk_shape)
        self.main_file = destination

    def _get_entries_stripped(self):
        return list(set([entry.split('[')[0] for entry in list(self.data.keys())]))

    def bib_citations(self, outfile=None):
        outfile = sys.stdout if outfile is None else outfile
        if self.metadata is None or self.metadata == {}:
            log.error("Metadata is not set for <%s> computational sequence" % self.root_name)
        outfile.write(
            'Computational Sequence <%s> bib: ' % self.root_name + self.metadata['featureset bib citation'] + '\n\n')
        outfile.write(
            'Dataset <%s> bib: ' % self.metadata["dataset name"] + self.metadata['dataset bib citation'] + '\n\n')

# some backward compatibility stuff.
computational_sequence.setData = computational_sequence.set_data
computational_sequence.setMetadata = computational_sequence.set_metadata
computational_sequence.completeAllMissingMetadata = computational_sequence.complete_all_missing_metadata


import sys
import torch as th
import torchvision.models as models
from mmkfeatures.video.resnext import resnet101
from torch import nn


class GlobalAvgPool(nn.Module):
    def __init__(self):
        super(GlobalAvgPool, self).__init__()

    def forward(self, x):
        return th.mean(x, dim=[-2, -1])

def get_model_p(model_type="2d",resnext101_model_path="model/resnext101.pth"):
    assert model_type in ['2d', '3d']
    if model_type == '2d':
        print('Loading 2D-ResNet-152 ...')
        model = models.resnet152(pretrained=True)
        model = nn.Sequential(*list(model.children())[:-2], GlobalAvgPool())
        model = model.cuda()
    else:
        print('Loading 3D-ResneXt-101 ...')
        model = resnet101(
            num_classes=400,
            shortcut_type='B',
            cardinality=32,
            sample_size=112,
            sample_duration=16,
            last_fc=False)
        model = model.cuda()
        model_data = th.load(resnext101_model_path)
        model.load_state_dict(model_data)

    model.eval()
    print('loaded')
    return model

def get_model(args):
    assert args.type in ['2d', '3d']
    if args.type == '2d':
        print('Loading 2D-ResNet-152 ...')
        model = models.resnet152(pretrained=True)
        model = nn.Sequential(*list(model.children())[:-2], GlobalAvgPool())
        model = model.cuda()
    else:
        print('Loading 3D-ResneXt-101 ...')
        model = resnet101(
            num_classes=400,
            shortcut_type='B',
            cardinality=32,
            sample_size=112,
            sample_duration=16,
            last_fc=False)
        model = model.cuda()
        model_data = th.load(args.resnext101_model_path)
        model.load_state_dict(model_data)

    model.eval()
    print('loaded')
    return model



import torch as th
from torch.utils.data.sampler import Sampler
import numpy as np

class RandomSequenceSampler(Sampler):

    def __init__(self, n_sample, seq_len):
        self.n_sample = n_sample
        self.seq_len = seq_len

    def _pad_ind(self, ind):
        zeros = np.zeros(self.seq_len - self.n_sample % self.seq_len)
        ind = np.concatenate((ind, zeros))
        return ind

    def __iter__(self):
        idx = np.arange(self.n_sample)
        if self.n_sample % self.seq_len != 0:
            idx = self._pad_ind(idx)
        idx = np.reshape(idx, (-1, self.seq_len))
        np.random.shuffle(idx)
        idx = np.reshape(idx, (-1))
        return iter(idx.astype(int))

    def __len__(self):
        return self.n_sample + (self.seq_len - self.n_sample % self.seq_len)



import librosa
import soundfile as sf
from tensorflow.keras.models import load_model
import pickle
import numpy as np
import librosa
import os

class SpeechEmotionToolkit:
    def __init__(self,model_path,model_para_path,VOICE_LEN=32000):
        self.VOICE_LEN = VOICE_LEN
        self.model_path=model_path
        self.model_para_path=model_para_path

    def normalizeVoiceLen(self,y, normalizedLen):
        nframes = len(y)
        y = np.reshape(y, [nframes, 1]).T
        # 归一化音频长度为2s,32000数据点
        if (nframes < normalizedLen):
            res = normalizedLen - nframes
            res_data = np.zeros([1, res], dtype=np.float32)
            y = np.reshape(y, [nframes, 1]).T
            y = np.c_[y, res_data]
        else:
            y = y[:, 0:normalizedLen]
        return y[0]

    def getNearestLen(self,framelength, sr):
        framesize = framelength * sr
        # 找到与当前framesize最接近的2的正整数次方
        nfftdict = {}
        lists = [32, 64, 128, 256, 512, 1024]
        for i in lists:
            nfftdict[i] = abs(framesize - i)
        sortlist = sorted(nfftdict.items(), key=lambda x: x[1])  # 按与当前framesize差值升序排列
        framesize = int(sortlist[0][0])  # 取最接近当前framesize的那个2的正整数次方值为新的framesize
        return framesize

    def load(self):
        # Load Model
        self.model = load_model(self.model_path)
        self.paradict = {}
        with open(self.model_para_path, 'rb') as f:
            paradict = pickle.load(f)
        self.DATA_MEAN = paradict['mean']
        self.DATA_STD = paradict['std']
        self.emotionDict = paradict['emotion']
        self.edr = dict([(i, t) for t, i in self.emotionDict.items()])

    def analyze_block(self,y,sr):

        N_FFT = self.getNearestLen(0.25, sr)

        y = self.normalizeVoiceLen(y, self.VOICE_LEN)  # 归一化长度

        mfcc_data = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=13, n_fft=N_FFT, hop_length=int(N_FFT / 4))
        feature = np.mean(mfcc_data, axis=0)
        feature = feature.reshape((126, 1))
        feature -= self.DATA_MEAN
        feature /= self.DATA_STD
        feature = feature.reshape((1, 126, 1))
        result = self.model.predict(feature)
        index = np.argmax(result, axis=1)[0]
        # print(self.edr[index])
        return self.edr[index]

    def analyze(self,wav_file):

        # Test

        # test_path = r'RawData/CASIA database/liuchanhg/angry/203.wav'

        y, sr = librosa.load(wav_file, sr=None)

        N_FFT = self.getNearestLen(0.25, sr)

        y = self.normalizeVoiceLen(y, self.VOICE_LEN)  # 归一化长度

        mfcc_data = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=13, n_fft=N_FFT, hop_length=int(N_FFT / 4))
        feature = np.mean(mfcc_data, axis=0)
        feature = feature.reshape((126, 1))
        feature -= self.DATA_MEAN
        feature /= self.DATA_STD
        feature = feature.reshape((1, 126, 1))
        result = self.model.predict(feature)
        index = np.argmax(result, axis=1)[0]
        # print(self.edr[index])
        return self.edr[index]

    def get_wav_blocks(self,audio_file, num_seconds_each_file):
        # file_name = "haodf.mp3"

        # First load the file
        audio, sr = librosa.load(audio_file)

        # Get number of samples for 2 seconds; replace 2 by any number
        buffer = num_seconds_each_file * sr

        samples_total = len(audio)
        samples_wrote = 0
        counter = 1

        list_blocks=[]
        while samples_wrote < samples_total:

            # check if the buffer is not exceeding total samples
            if buffer > (samples_total - samples_wrote):
                buffer = samples_total - samples_wrote

            block = audio[samples_wrote: (samples_wrote + buffer)]
            list_blocks.append((block,sr))

            counter += 1
            samples_wrote += buffer
        return list_blocks

    def get_wav_files(self,audio_file, num_seconds_each_file, temp_folder):
        # file_name = "haodf.mp3"

        # First load the file
        audio, sr = librosa.load(audio_file)

        # Get number of samples for 2 seconds; replace 2 by any number
        buffer = num_seconds_each_file * sr

        samples_total = len(audio)
        samples_wrote = 0
        counter = 1
        list_files = []
        while samples_wrote < samples_total:

            # check if the buffer is not exceeding total samples
            if buffer > (samples_total - samples_wrote):
                buffer = samples_total - samples_wrote

            block = audio[samples_wrote: (samples_wrote + buffer)]
            out_filename = temp_folder + "/split_" + str(counter) + "_" + audio_file
            out_filename = out_filename.replace(".mp3", ".wav")
            list_files.append(out_filename)
            # Write 2 second segment
            # librosa.output.write_wav(out_filename, block, sr)
            sf.write(out_filename, block, sr, 'PCM_24')
            counter += 1
            samples_wrote += buffer
        return list_files

    def get_emotion_list(self,audio_file,num_sec_each_file,temp_folder):
        list_files = self.get_wav_files(audio_file, num_sec_each_file,temp_folder=temp_folder)
        list_ts=[]
        list_emo=[]
        count=0
        for file in list_files:
            emotion = self.analyze(file)
            list_emo.append(emotion)
            list_ts.append([count*num_sec_each_file,(count+1)*num_sec_each_file])
            os.remove(file)
            count+=1
        return list_emo,list_ts

    def get_emotion_list_by_blocks(self, audio_file, num_sec_each_file):
        list_blocks = self.get_wav_blocks(audio_file, num_sec_each_file)
        list_ts = []
        list_emo = []
        count = 0
        for block_info in list_blocks:
            emotion = self.analyze_block(block_info[0],block_info[1])
            list_emo.append(emotion)
            list_ts.append([count * num_sec_each_file, (count + 1) * num_sec_each_file])
            count += 1
        return list_emo,list_ts

def example_use():
    speech_kit = SpeechEmotionToolkit(model_path='speech_emotion/speech_mfcc_model.h5', model_para_path='speech_emotion/mfcc_model_para_dict.pkl')
    speech_kit.load()
    list_emo, list_timestamp = speech_kit.get_emotion_list_by_blocks(audio_file="speech_emotion/haodf.mp3", num_sec_each_file=5)

    print("Time interval\tEmotion")
    for idx, e in enumerate(list_emo):
        print(list_timestamp[idx], "\t", e)

# example_use()





import sys
import pickle # For serializing data
import os.path # For checking whether a file exist
from nltk.stem import PorterStemmer as ps # For stemming and word tokenization

class PositionalIndex:
    def __init__(self):
        pass

    _STOP_WORDS = frozenset([
        'a', 'about', 'above', 'above', 'across', 'after', 'afterwards', 'again',
        'against', 'all', 'almost', 'alone', 'along', 'already', 'also', 'although',
        'always', 'am', 'among', 'amongst', 'amoungst', 'amount', 'an', 'and', 'another',
        'any', 'anyhow', 'anyone', 'anything', 'anyway', 'anywhere', 'are', 'around', 'as',
        'at', 'back', 'be', 'became', 'because', 'become', 'becomes', 'becoming', 'been',
        'before', 'beforehand', 'behind', 'being', 'below', 'beside', 'besides',
        'between', 'beyond', 'bill', 'both', 'bottom', 'but', 'by', 'call', 'can',
        'cannot', 'cant', 'co', 'con', 'could', 'couldnt', 'cry', 'de', 'describe',
        'detail', 'do', 'done', 'down', 'due', 'during', 'each', 'eg', 'eight',
        'either', 'eleven', 'else', 'elsewhere', 'empty', 'enough', 'etc', 'even',
        'ever', 'every', 'everyone', 'everything', 'everywhere', 'except', 'few',
        'fifteen', 'fify', 'fill', 'find', 'fire', 'first', 'five', 'for', 'former',
        'formerly', 'forty', 'found', 'four', 'from', 'front', 'full', 'further', 'get',
        'give', 'go', 'had', 'has', 'hasnt', 'have', 'he', 'hence', 'her', 'here',
        'hereafter', 'hereby', 'herein', 'hereupon', 'hers', 'herself', 'him',
        'himself', 'his', 'how', 'however', 'hundred', 'ie', 'if', 'in', 'inc',
        'indeed', 'interest', 'into', 'is', 'it', 'its', 'itself', 'keep', 'last',
        'latter', 'latterly', 'least', 'less', 'ltd', 'made', 'many', 'may', 'me',
        'meanwhile', 'might', 'mill', 'mine', 'more', 'moreover', 'most', 'mostly',
        'move', 'much', 'must', 'my', 'myself', 'name', 'namely', 'neither', 'never',
        'nevertheless', 'next', 'nine', 'no', 'nobody', 'none', 'noone', 'nor', 'not',
        'nothing', 'now', 'nowhere', 'of', 'off', 'often', 'on', 'once', 'one', 'only',
        'onto', 'or', 'other', 'others', 'otherwise', 'our', 'ours', 'ourselves', 'out',
        'over', 'own', 'part', 'per', 'perhaps', 'please', 'put', 'rather', 're', 'same',
        'see', 'seem', 'seemed', 'seeming', 'seems', 'serious', 'several', 'she',
        'should', 'show', 'side', 'since', 'sincere', 'six', 'sixty', 'so', 'some',
        'somehow', 'someone', 'something', 'sometime', 'sometimes', 'somewhere',
        'still', 'such', 'system', 'take', 'ten', 'than', 'that', 'the', 'their',
        'them', 'themselves', 'then', 'thence', 'there', 'thereafter', 'thereby',
        'therefore', 'therein', 'thereupon', 'these', 'they', 'thickv', 'thin', 'third',
        'this', 'those', 'though', 'three', 'through', 'throughout', 'thru', 'thus',
        'to', 'together', 'too', 'top', 'toward', 'towards', 'twelve', 'twenty', 'two',
        'un', 'under', 'until', 'up', 'upon', 'us', 'very', 'via', 'was', 'we', 'well',
        'were', 'what', 'whatever', 'when', 'whence', 'whenever', 'where', 'whereafter',
        'whereas', 'whereby', 'wherein', 'whereupon', 'wherever', 'whether', 'which',
        'while', 'whither', 'who', 'whoever', 'whole', 'whom', 'whose', 'why', 'will',
        'with', 'within', 'without', 'would', 'yet', 'you', 'your', 'yours', 'yourself',
        'yourselves', 'the'])
    # Removes most special characters and caps
    def preprocess(self,data):
        for p in "!.,:@#$%^&?<>*()[}{]-=;/\"\\\t\n":
            if p in '\n;?:!.,.':
                data = data.replace(p, ' ')
            else:
                data = data.replace(p, '')
        if " " not in data:
            return data
        words=data.split(" ")
        list_w=[]
        for w in words:
            if w.lower() not in self._STOP_WORDS:
                list_w.append(w)
        return " ".join(list_w).lower()

    # For each file, opens and adds it to the hashmap
    def createPositionalIndex(self,data_list):
        index = {}
        for i in range(len(data_list)):
            key = data_list[i][0]
            if type(data_list[i][1])!=str:
                text = data_list[i][1].decode(encoding='UTF-8',errors='ignore')
            else:
                text=data_list[i][1]
            # with open(files[i],encoding='utf-8') as f:
            #    doc = [a for a in preprocess(f.read()).split(' ') if a != ""]
            # print(text)
            words=self.preprocess(text).split(' ')
            # print(words)
            doc = [a for a in words if a != ""]
            for idx, word in enumerate(doc):
                stemmed = ps().stem(word)
                if not stemmed in index:
                    index[stemmed] = [(key, idx)]
                else:
                    index[stemmed].append((key, idx))
        return index

    # shows a preview based on the positions and the how
    # much text to show around the data found
    def showPreview(self,data_list, positions, radius):
        for i, (doc_id, word_index) in enumerate(positions):
            for data in data_list:
                if data[0] == doc_id:
                    wordArr = [a for a in self.preprocess(data[1]).split(' ') if a != ""]
                    result = " ".join(wordArr[word_index - radius:word_index + radius])
                    print(str(i + 1) + ": ..." + result + "... " + data[0])
                    # print(data[0],":",data[1])
        print()

    def create(self,plain_text_index_file,positional_index_file):
        list_id_text = pickle.load(open(plain_text_index_file, 'rb'))
        print("Processing and serializing data for future use...")
        pi = self.createPositionalIndex(list_id_text)
        with open(positional_index_file, "wb") as f:
            pickle.dump(pi, f)

    def create_with_data(self,list_id_text,positional_index_file):

        print("Processing and serializing data for future use...")
        pi = self.createPositionalIndex(list_id_text)
        with open(positional_index_file, "wb") as f:
            pickle.dump(pi, f)

    def positional_indexing_search_2words(self,positional_index_file,word1, word2):
        pi=pickle.load(open(positional_index_file,"rb"))
        word1 = ps().stem(self.preprocess(word1).replace(' ', ''))
        word2 = ps().stem(self.preprocess(word2).replace(' ', ''))
        matches = []
        for doc1, index1 in pi[word1]:
            for doc2, index2 in pi[word2]:
                if doc1 != doc2: continue
                if index1 == (index2 - 1): matches.append((doc1, index1))
        list_ids=[]
        for i, (doc_id, word_index) in enumerate(matches):
            list_ids.append(doc_id)
        return list_ids

        # showPreview(list_id_text, matches, 5)

    def positional_indexing_search_2words_num(self,positional_index_file,word1, word2, num):
        pi = pickle.load(open(positional_index_file, "rb"))
        matches = []
        word1 = ps().stem(self.preprocess(word1).replace(' ', ''))
        word2 = ps().stem(self.preprocess(word2).replace(' ', ''))
        print("Searching... \n")
        rad = int(num)
        for doc1, index1 in pi[word1]:
            for doc2, index2 in pi[word2]:
                if doc1 != doc2: continue
                abs_pos = abs(index1 - index2)
                # when abs_pos is 0, the word is itself
                if abs_pos <= rad and abs_pos != 0: matches.append((doc1, index1))
        # showPreview(matches, 5 if rad <= 5 else rad)
        return matches

    def test(self,index_file):
        # test search example
        import time
        queries = ['gram-negative bacteria', 'Fungal infection', 'purulent exudate']

        start_time = time.time()
        for q in queries:
            ws = q.split(" ")
            self.positional_indexing_search_2words(index_file,ws[0], ws[1])

        end_time = time.time()

        print("time cost: ", round((end_time - start_time) * 1000, 2), "ms")





import numpy as np
from sklearn.cluster import KMeans
from scipy.spatial.distance import cdist


class ClusteringModel(object):
    def __init__(self, min_k: int = 1, max_k: int = 1, step: int = 1, threshold: float = 0.85):
        self.data = None
        self.selected_model = None
        self.models = []
        self.distances = []
        
        self.k = 0
        self.threshold = threshold
        self.k_grid = list(range(min_k, max_k, step))
    
    def add(self, descriptors):
        if self.data is not None:
            self.data = np.concatenate((self.data, descriptors), axis=0)
        else:
            self.data = descriptors
    
    def cluster(self) -> np.ndarray:
        for k in self.k_grid:
            print("Clustering for k: {}".format(k))
            self.cluster_and_evaluate_by_k(k)
        
        self.find_best_model()
        
        return self.selected_model.cluster_centers_
    
    def cluster_and_evaluate_by_k(self, k: int):
        model = KMeans(n_clusters=k).fit(self.data)
        distance = self.evaluate(model)
        
        self.models.append(model)
        self.distances.append(distance)
    
    def evaluate(self, model: KMeans) -> int:
        total_distance = 0
        predicted_clusters = model.fit_predict(self.data)
        
        for sample, cluster_idx in zip(self.data, predicted_clusters):
            centroid = model.cluster_centers_[cluster_idx]
            total_distance += ClusteringModel.get_distance(sample, centroid)
        
        return total_distance
    
    def find_best_model(self):
        
        for idx, dist in enumerate(self.distances[:-1]):
            ratio = dist / self.distances[idx + 1]
            
            if ratio > self.threshold:
                self.k = self.k_grid[idx]
                self.selected_model = self.models[idx]
                break
    
    @staticmethod
    def get_distance(sample: np.ndarray, centroid: np.ndarray) -> int:
        return cdist(np.array([sample]), np.array([centroid]), 'sqeuclidean')[0, 0]



import json
import torch

torch.cuda.empty_cache()
torch.cuda.set_per_process_memory_fraction(1.0, 0)
torch.cuda.memory_summary(device=None, abbreviated=False)
import torch.backends.cudnn as cudnn
cudnn.benchmark = True

from torch.utils.data import DataLoader
import random
import os
import argparse
import time
from pathlib import Path
import sys
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

# Import models
path = os.path.dirname(os.path.abspath(__file__))
path = Path(path)
src_path = path.parent
print("Add models path to sys paths - ", src_path)
sys.path.append(str(src_path) + '/models/char_embedding/')
from mmkfeatures.models.image_text_fusion.char_embedding.char_cnn_rnn import CharCnnRnn
from mmkfeatures.models.image_text_fusion.char_embedding.multi_model_dataset import MultimodalDataset

def train(json_path, output_path, learning_rate=0.001, img_tag='encod_64x64_path',
          learning_rate_decay=0.1, batch_size=20, epochs=20,
          symmetric=True, num_workers=0, model_type='fixed_gru', rnn_type='cvpr'):
    ensure_folder(output_path)
    # Train model
    device = 'cuda' if torch.cuda.is_available() else 'cpu'

    # Create datasets loader
    train_data, test_data = get_datasets(json_path, device, 0.8, img_tag)
    print("Creating dataloaders ...")
    train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=False,
                              num_workers=num_workers)
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False,
                             num_workers=num_workers)
    print("Dataloaders created!")
    # create Char-CNN-RNN model
    model = CharCnnRnn(model_type, rnn_type)
    print(f"CharCnnRnn created device:{device}")
    optim = torch.optim.RMSprop(model.parameters(), lr=learning_rate)
    # Adjust lr in progress
    sched = torch.optim.lr_scheduler.ExponentialLR(optim, learning_rate_decay)

    # torch.cuda.empty_cache()
    # torch.set_grad_enabled(False)

    model.to(device)
    model.train()
    ###################
    # train the model #
    ###################

    for epoch in range(epochs):
        for i, data in enumerate(train_loader):
            img = data['img']
            txt = data['text']
            feat_img = img.view(img.size(0), -1)
            feat_txt = model(txt)
            loss1, acc1 = sje_loss(feat_txt, feat_img)
            loss2 = acc2 = 0
            if symmetric:
                loss2, acc2 = sje_loss(feat_img, feat_txt)
            loss = loss1 + loss2
            model.zero_grad()
            loss.backward()
            optim.step()


        print(f"Epoch: {epoch} Loss: {loss}")
        sched.step()
    ######################
    # evaluate the model #
    ######################
    print("evaling...")
    torch.cuda.empty_cache()
    torch.set_grad_enabled(False)
    model.eval()
    eval_loss = 0
    for data in test_loader:

        img = data['img']
        txt = data['text']
        feat_img = img.view(img.size(0), -1)
        feat_txt = model(txt)
        loss1, acc1 = sje_loss(feat_txt, feat_img)
        loss2 = acc2 = 0
        if symmetric:
            loss2, acc2 = sje_loss(feat_img, feat_txt)
        loss = loss1 + loss2
        # update evaluation loss
        eval_loss += loss
    eval_loss = eval_loss / len(test_loader)
    print(f"Evaluation loss: {eval_loss}")
    # Save model
    # model_root = output_path + '/char_embedding' + str(time_msec()) + '.pt'
    model_root = output_path + '/char_embedding.pt'
    print(f"Saving trained model to: {model_root}")
    # torch.cuda.empty_cache()

    model_state_dict=model.state_dict()
    torch.save(model_state_dict, model_root)


def ensure_folder(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)


def time_msec():
    return int(round(time.time() * 1000))


def get_datasets(dataset_json_path, device, split=0.8, img_tag='encod_64x64_path'):
    # Load data and create train and test datasets loaders
    data_folder = os.path.dirname(dataset_json_path)
    items = []
    with open(dataset_json_path) as f:
        dataset_json = json.load(f)
        # dataset_json=random.sample(dataset_json,60000)
        print(f"Load dataset size: {len(dataset_json)}")
        for item in dataset_json:
            items.append({'img': item[img_tag], 'text': item['text']})
        # Split train and test datasets
        train_size = int(len(items) * split)
        train_indx = random.sample(range(len(items)), k=train_size)
        indx = [i for i in range(len(items))]
        test_indx = list(set(indx) - set(train_indx))
        train_dataset = [items[i] for i in train_indx]
        test_dataset = [items[i] for i in test_indx]
        print(f"Train : {len(train_dataset)} Test: {len(test_dataset)}")

        print(f"Creating MultimodalDatasets from: {data_folder}")
        train_data = MultimodalDataset(train_dataset, data_folder, device)
        test_data = MultimodalDataset(test_dataset, data_folder, device)

        print("MultimodalDatasets created!")

        return train_data, test_data


def sje_loss(feat1, feat2):
    # Structured Joint Embedding Loss
    # similarity score matrix (rows: fixed feat2, columns: fixed feat1)
    # print(feat2.shape,feat1.t().shape)
    scores = torch.matmul(feat2, feat1.t())  # (B, B)
    # diagonal: matching pairs
    diagonal = scores.diag().view(scores.size(0), 1)  # (B, 1)
    # repeat diagonal scores on rows
    diagonal = diagonal.expand_as(scores)  # (B, B)
    # calculate costs
    cost = (1 + scores - diagonal).clamp(min=0)  # (B, B)
    # clear diagonals (matching pairs are not used in loss computation)
    # cost[torch.eye(cost.size(0)).bool()] = 0 # (B, B) for torch==1.2.0
    cost[torch.eye(cost.size(0), dtype=torch.bool)] = 0  # (B, B)
    # sum and average costs
    denom = cost.size(0) * cost.size(1)
    loss = cost.sum() / denom

    # batch accuracy
    max_ids = torch.argmax(scores, dim=1)
    ground_truths = torch.LongTensor(range(scores.size(0))).to(feat1.device)
    num_correct = (max_ids == ground_truths).sum().float()
    accuracy = 100 * num_correct / cost.size(0)

    return loss, accuracy


def main():
    '''
train.py
../../dataset/images_data.json
../../output
0.001
100
fixed_gru
cvpr
encod_64x64_path
    '''
    parser = argparse.ArgumentParser(description='Train embedding model on dataset.')
    parser.add_argument('json_path', metavar='json_path', type=str,
                        help='Source folder full path.')
    parser.add_argument('output', metavar='output', type=str,
                        help='Destination folder full path.')
    parser.add_argument('lr', metavar='lt', type=str,
                        help='Learning rate')
    parser.add_argument('epochs', metavar='epochs', type=str,
                        help='Number of training epochs.')
    parser.add_argument('model_type', metavar='model_type', type=str,
                        help='def fixed_gru|fixed_rnn')
    parser.add_argument('rnn_type', metavar='rnn_type', type=str,
                        help='def cvpr|icml')
    parser.add_argument('img_tag', metavar='img_tag', type=str,
                        help='def img_64x64_path|img_256x256_path')

    args = parser.parse_args()
    json_path = args.json_path
    output_folder = args.output
    lr = args.lr
    epochs = args.epochs
    model_type = args.model_type
    rnn_type = args.rnn_type
    img_tag = args.img_tag
    train(json_path=json_path, output_path=output_folder, learning_rate=float(lr), epochs=int(epochs), model_type=model_type, rnn_type=rnn_type, img_tag=img_tag)


if __name__ == '__main__':
    print("<<<<<<<<<<<<<<< Start training >>>>>>>>>>>>>>>>>")
    main()



#CMU Multimodal SDK, CMU Multimodal Model SDK

#Multimodal Language Analysis with Recurrent Multistage Fusion, Paul Pu Liang, Ziyin Liu, Amir Zadeh, Louis-Philippe Morency - https://arxiv.org/abs/1808.03920 

#in_dimensions: the list of dimensionalities of each modality 

#cell_size: lstm cell size

#in_modalities: is a list of inputs from each modality - the first dimension of all the modality inputs must be the same, it will be the batch size. The second dimension is the feature dimension. There are a total of n modalities.

#steps: number of iterations for the recurrent fusion

import torch
import time
from torch import nn
import torch.nn.functional as F
from six.moves import reduce

class RecurrentFusion(nn.Module):

	def __init__(self,in_dimensions,cell_size):
		super(RecurrentFusion, self).__init__()
		self.in_dimensions=in_dimensions
		self.cell_size=cell_size
		self.model=nn.LSTM(sum(in_dimensions),cell_size)
	def __call__(self,in_modalities,steps=1):
		return self.fusion(in_modalities,steps)

	def fusion(self,in_modalities,steps=1):
		bs=in_modalities[0].shape[0]
		model_input=torch.cat(in_modalities,dim=1).view(1,bs,-1).repeat([steps,1,1])
		hidden,cell = (torch.zeros(1, bs, self.cell_size),torch.zeros(1, bs, self.cell_size))
		for i in range(steps):
			outputs,last_states=self.model(model_input,[hidden,cell])
		return outputs,last_states[0],last_states[1]
		
	def forward(self, x):
		print("Not yet implemented for nn.Sequential")
		exit(-1)

if __name__=="__main__":
	print("This is a module and hence cannot be called directly ...")
	print("A toy sample will now run ...")
	
	from torch.autograd import Variable
	import torch.nn.functional as F
	import numpy

	inputx=Variable(torch.Tensor(numpy.zeros([32,40])),requires_grad=True)
	inputy=Variable(torch.Tensor(numpy.array(numpy.zeros([32,12]))),requires_grad=True)
	inputz=Variable(torch.Tensor(numpy.array(numpy.zeros([32,20]))),requires_grad=True)
	modalities=[inputx,inputy,inputz]
	
	fmodel=RecurrentFusion([40,12,20],100)
	
	out=fmodel(modalities,steps=5)

	print("Output")
	print(out[0])
	print("Toy sample finished ...")





import os
from typing import List, Dict

import cv2
import numpy as np

from mmkfeatures.image.extractor import Extractor
from  mmkfeatures.image.clutering_model import ClusteringModel


class BoWExtractor(Extractor):
    
    def __init__(self, base_route: str, method: str, k: int = 0, min_k: int = 0, max_k: int = 0, step: int = 0,
                 threshold: float = 0.85, cluster_mode: str = 'manual', size: int = 224, batch_size: int = 128,
                 color_code: int = cv2.COLOR_BGR2RGB):
        
        super().__init__(base_route=base_route, size=size, batch_size=batch_size)
        
        self.k: int = k
        self.image_keypoints: Dict = {}
        
        self.detector = None
        self.bow_kmeans = None
        self.bow_extractor = None
        self.__empty_histogram: List[float] = [0.0] * self.k
        self.__color_code: int = color_code
        
        self._set_detector(method)
        self._set_clustering_trainer(mode=cluster_mode, min_k=min_k, max_k=max_k, step=step, threshold=threshold)
    
    def setup(self):
        for filename, category in zip(self.directory_iterator.filenames, self.directory_iterator.classes):
            self._extract_and_add_image(os.path.join(self.base_route, filename))
            self._update_counter()
    
    def fit(self):
        vocabulary = self.bow_kmeans.cluster()
        self.bow_extractor = cv2.BOWImgDescriptorExtractor(self.detector, cv2.BFMatcher(cv2.NORM_L2))
        self.bow_extractor.setVocabulary(vocabulary)
    
    def export(self, export_route: str):
        vocabulary: np.ndarray = self.bow_extractor.getVocabulary()
        
        if '.npy' in export_route:
            np.save(export_route, vocabulary)
        elif '.npz' in export_route:
            np.savez(export_route, vocabulary)
        else:
            raise NotImplementedError()
    
    def load(self, import_route: str):
        vocabulary: np.ndarray = np.load(import_route)
        self.bow_extractor = cv2.BOWImgDescriptorExtractor(self.detector, cv2.BFMatcher(cv2.NORM_L2))
        self.bow_extractor.setVocabulary(vocabulary)
    
    def extract(self, image_route: str):
        image = self._read_image(image_route)
        # print("image_keypoints",self.image_keypoints.keys())
        keypoints = self.image_keypoints[image_route]
        if keypoints is None or len(keypoints) == 0:
            return self.__empty_histogram
        
        return self.bow_extractor.compute(image, keypoints)[0]
    
    def _set_detector(self, method: str):
        if method.lower() == "kaze":
            self.detector = cv2.KAZE_create()
        elif method.lower() == "sift":
            self.detector = cv2.xfeatures2d.SIFT_create()
        elif method.lower() == "surf":
            self.detector = cv2.xfeatures2d.SURF_create()
        else:
            raise Exception("Non implemented method")
    
    def _set_clustering_trainer(self, mode: str, min_k: int = 0, max_k: int = 0, step: int = 0,
                                threshold: float = 0.85):
        if mode == 'manual':
            self.bow_kmeans = cv2.BOWKMeansTrainer(self.k)
        elif mode == 'auto':
            self.bow_kmeans = ClusteringModel(min_k, max_k, step, threshold)
    
    def _extract_and_add_image(self, image_route: str):
        keypoints, descriptors = self._process_image(image_route)
        self._add_image(image_route, keypoints, descriptors)
    
    def _process_image(self, image_route: str) -> (List[cv2.KeyPoint], np.ndarray):
        return self.detector.detectAndCompute(self._read_image(image_route), None)
    
    def _read_image(self, image_route: str):
        return cv2.cvtColor(cv2.imread(image_route), self.__color_code)
    
    def _add_image(self, image_route: str, keypoints: List[cv2.KeyPoint], descriptors: np.ndarray):
        if descriptors is not None and len(descriptors) > 0:
            self.bow_kmeans.add(descriptors)
        image_route=image_route.replace("\\","/")
        self.image_keypoints[image_route] = keypoints
    
    def _find_features_size(self) -> int:
        return self.k



from mmkfeatures.fusion.computational_sequence  import  computational_sequence
from mmkfeatures.fusion import log
import sys
import numpy
import time
from tqdm import tqdm
import os

#specified for numerical inconsistencies within floating points - if abs l1 distance two numbers is less than this, then they are the same. 
#only use for interval comparison.
epsilon=10e-6

class mmdataset:

	def __init__(self,recipe,destination=None):
		self.computational_sequences={}

		if type(recipe) is str:
			if os.path.isdir(recipe) is False:
				log.error("Dataset folder does not exist ...",error=True)

			from os import listdir
			from os.path import isfile, join
			computational_sequence_list = [f for f in listdir(recipe) if isfile(join(recipe, f)) and f[-4:]=='.csd']
			for computational_sequence_fname in computational_sequence_list:
				this_sequence=computational_sequence(join(recipe,computational_sequence_fname))
				self.computational_sequences[this_sequence.metadata["root name"]]=this_sequence

		elif type(recipe) is dict:
			for entry, address in recipe.items():
				self.computational_sequences[entry]=computational_sequence(address,destination)

		elif type(recipe) is list:
			#TODO: computational sequence with uuid
			pass

		else:
			log.error("Cannot create a dataset with the given recipe. Exiting ...!",error=True)

		if len(self.computational_sequences.keys())==0:
			log.error("Dataset failed to initialize ...", error=True)

		log.success("Dataset initialized successfully ... ")

	def __getitem__(self,key):
		if key not in list(self.computational_sequences.keys()):
			log.error("Computational sequence does not exist ...",error=True)
		return self.computational_sequences[key]

	def keys(self):
		return self.computational_sequences.keys()

	def add_computational_sequences(self,recipe,destination):
		for entry, address in recipe.items():
			if entry in self.computational_sequences:
				log.error("Dataset already contains <%s> computational sequence ..."%entry)
			self.computational_sequences[entry]=computational_sequence(address,destination)

	def bib_citations(self,outfile=None):
		outfile=sys.stdout if outfile is None else outfile
		sdkbib='@article{zadeh2018multi, title={Multi-attention recurrent network for human communication comprehension}, author={Zadeh, Amir and Liang, Paul Pu and Poria, Soujanya and Vij, Prateek and Cambria, Erik and Morency, Louis-Philippe}, journal={arXiv preprint arXiv:1802.00923}, year={2018}}'
		outfile.write('mmsdk bib: '+sdkbib+'\n\n')
		for entry,compseq in self.computational_sequences.items():
			compseq.bib_citations(outfile)

	def unify(self,active=True):
		log.status("Unify was called ...")

		all_vidids={}
		violators=[]
		
		all_keys={}
		for seq_key in list(self.computational_sequences.keys()):
			all_keys[seq_key]=[vidid.split("[")[0] for vidid in self.computational_sequences[seq_key].data.keys()]
		
		valids=set.intersection(*[set(all_keys[x]) for x in all_keys])
		violators=set()
		for seq_key in list(self.computational_sequences.keys()):
			violators=violators.union(set([vidid.split("[")[0] for vidid in self.computational_sequences[seq_key].data.keys()])-valids)
		
		if len(violators) >0 :
			for violator in violators:
				log.error("%s entry is not shared among all sequences, removing it ..."%violator,error=False)
				if active==True:
					self.remove_id(violator,purge=True)
		if active==False and len(violators)>0:
			log.error("%d violators remain, alignment will fail if called ..."%len(violators),error=True)

		log.success("Unify completed ...")

	def hard_unify(self,active=True):
		log.status("Hard unify was called ...")

		all_vidids={}
		violators=[]
		
		all_keys={}
		for seq_key in list(self.computational_sequences.keys()):
			all_keys[seq_key]=[vidid for vidid in self.computational_sequences[seq_key].data.keys()]
		
		valids=set.intersection(*[set(all_keys[x]) for x in all_keys])
		for seq_key in list(self.computational_sequences.keys()):
			hard_unify_compatible=all(["[" in vidid for vidid in self.computational_sequences[seq_key].data.keys()])
			if hard_unify_compatible is False:
				log.error("Hard unify can only be done on aligned computational sequences, %s violated this ... Exiting ..."%seq_key)
			violators=set([vidid for vidid in self.computational_sequences[seq_key].data.keys()])-valids
			for violator in violators:
				if active==True:
					log.error("%s entry is not shared among all sequences, removing it ..."%violator,error=False)
					self[seq_key]._remove_id(violator,purge=False)

			if active==False and len(violators)>0:
				log.error("%d violators remain, alignment will fail if called ..."%len(violators),error=True)
		
		log.success("Hard unify completed ...")


	
	def remove_id(self,entry_id,purge=False):
		for key,compseq in self.computational_sequences.items():
			compseq._remove_id(entry_id,purge=purge)

	def align(self,reference,collapse_functions=None,replace=True):
		aligned_output={}

		for sequence_name in self.computational_sequences.keys():
			aligned_output[sequence_name]={}
		if reference not in self.computational_sequences.keys():
			log.error("Computational sequence <%s> does not exist in dataset"%reference,error=True)
		refseq=self.computational_sequences[reference].data
		#unifying the dataset, removing any entries that are not in the reference computational sequence
		self.unify()

		#building the relevant entries to the reference - what we do in this section is simply removing all the [] from the entry ids and populating them into a new dictionary
		log.status("Pre-alignment based on <%s> computational sequence started ..."%reference)
		relevant_entries=self.__get_relevant_entries(reference)
		log.status("Alignment starting ...")

		pbar = log.progress_bar(total=len(refseq.keys()),unit=" Computational Sequence Entries",leave=False)
		pbar.set_description("Overall Progress")
		for entry_key in list(refseq.keys()):
			pbar_small=log.progress_bar(total=refseq[entry_key]['intervals'].shape[0],unit=" Segments",leave=False)
			pbar_small.set_description("Aligning %s"%entry_key)
			for i in range(refseq[entry_key]['intervals'].shape[0]):
				#interval for the reference sequence
				ref_time=refseq[entry_key]['intervals'][i,:]
				#we drop zero or very small sequence lengths - no align for those
				if (abs(ref_time[0]-ref_time[1])<epsilon):
					pbar_small.update(1)
					continue

				#aligning all sequences (including ref sequence) to ref sequence
				for otherseq_key in list(self.computational_sequences.keys()):
					if otherseq_key != reference:
						intersects,intersects_features=self.__intersect_and_copy(ref_time,relevant_entries[otherseq_key][entry_key],epsilon)
					else:
						intersects,intersects_features=refseq[entry_key]['intervals'][i,:][None,:],refseq[entry_key]['features'][i,:][None,:]
					#there were no intersections between reference and subject computational sequences for the entry
					if intersects.shape[0] == 0:
						continue
					#collapsing according to the provided functions
					if type(collapse_functions) is list:
						intersects,intersects_features=self.__collapse(intersects,intersects_features,collapse_functions)
					if(intersects.shape[0]!=intersects_features.shape[0]):
						log.error("Dimension mismatch between intervals and features when aligning <%s> computational sequences to <%s> computational sequence"%(otherseq_key,reference),error=True)
					aligned_output[otherseq_key][entry_key+"[%d]"%i]={}
					aligned_output[otherseq_key][entry_key+"[%d]"%i]["intervals"]=intersects
					aligned_output[otherseq_key][entry_key+"[%d]"%i]["features"]=intersects_features
				pbar_small.update(1)
			pbar_small.close()
			pbar.update(1)
		pbar.close()
		log.success("Alignment to <%s> complete."%reference)
		if replace is True:
			log.status("Replacing dataset content with aligned computational sequences")
			self.__set_computational_sequences(aligned_output)
			return None
		else:
			log.status("Creating new dataset with aligned computational sequences")
			newdataset=mmdataset({})
			newdataset.__set_computational_sequences(aligned_output,metadata_copy=False)
			return newdataset

	def __collapse(self,intervals,features,functions):
		#we simply collapse the intervals to (1,2) matrix
		new_interval=numpy.array([[intervals.min(),intervals.max()]])
		try:
			new_features=numpy.concatenate([function(intervals,features) for function in functions],axis=0)
			if len(new_features.shape)==1:
				new_features=new_features[None,:]
		except:
			log.error("Cannot collapse given the set of functions. Please check for exceptions in your functions ...", error=True)
		return new_interval,new_features

	#TODO: This is not passive-align safe
	def revert(self,replace=True):
		reverted_dataset={x:{} for x in self.keys()}
		log.status("Revert was called ...")
		if len(self.keys())==0:
			log.error("The dataset contains no computational sequences ... Exiting!",error=True)

		self.hard_unify()
		all_keys=list(self[list(self.keys())[0]].keys())

		if len(all_keys)==0:
			log.error("No entries in computational sequences or removed during unify ... Exiting!")

		unique_unnumbered_entries={}

		for key in all_keys:
			if key.split('[')[0] not in unique_unnumbered_entries:
				unique_unnumbered_entries[key.split('[')[0]]=[]
			unique_unnumbered_entries[key.split('[')[0]].append(int(key.split('[')[1][:-1]))

		pbar = tqdm(total=len(unique_unnumbered_entries.keys()),unit=" Unique Sequence Entries",leave=False)
		pbar.set_description("Reversion Progress")
		for key in unique_unnumbered_entries.keys():
			unique_unnumbered_entries[key].sort()
			for cs_key in reverted_dataset.keys():
				intervals=numpy.concatenate([self[cs_key][str('%s[%d]'%(key,i))]["intervals"] for i in unique_unnumbered_entries[key]],axis=0)
				features=numpy.concatenate([self[cs_key][str('%s[%d]'%(key,i))]["features"] for i in unique_unnumbered_entries[key]],axis=0)
				reverted_dataset[cs_key][key]={"intervals":intervals,"features":features}
			pbar.update(1)
		pbar.close()
		log.success("Reversion completed ...")
		if replace is True:
			log.status("Replacing dataset content with reverted computational sequences")
			self.__set_computational_sequences(reverted_dataset)
			return None
		else:
			log.status("Creating new dataset with reverted computational sequences")
			newdataset=mmdataset({})
			newdataset.__set_computational_sequences(reverted_dataset,metadata_copy=False)
			return newdataset	

	def impute(self,ref_key,imputation_fn=numpy.zeros):
		log.status("Imputation called with function: %s ..."%str(imputation_fn.__name__))
		other_keys=list(self.keys())
		other_keys.remove(ref_key)
		other_keys_dims={x:list(self[x][list(self[x].keys())[0]]["features"].shape[1:]) for x in other_keys}
		pbar = tqdm(total=len(self[ref_key].keys()),unit=" Reference Computational Sequence Entries",leave=False)
		pbar.set_description("Imputation Progress")
		for seg_key in self[ref_key].keys():
			for other_key in other_keys:
				try:
					self[other_key][seg_key]
				except:
					num_entries=self[ref_key][seg_key]["intervals"].shape[0]
					self[other_key][seg_key]={"intervals":self[ref_key][seg_key]["intervals"],"features":imputation_fn([num_entries]+other_keys_dims[other_key])}
					#log.status("Imputing %s,%s with function %s"%(other_key,seg_key,str(imputation_fn.__name__)))
			pbar.update(1)
		pbar.close()
		log.success("Imputation completed ...")


	#setting the computational sequences in the dataset based on a given new_computational_sequence_data - may copy the metadata if there is already one
	def __set_computational_sequences(self,new_computational_sequences_data,metadata_copy=True):

		#getting the old metadata from the sequence before replacing it. Even if this is a new computational sequence this will not cause an issue since old_metadat will just be empty
		old_metadata={m:self.computational_sequences[m].metadata for m in list(self.computational_sequences.keys())}
		self.computational_sequences={}
		for sequence_name in list(new_computational_sequences_data.keys()):
			self.computational_sequences[sequence_name]=computational_sequence(sequence_name)
			self.computational_sequences[sequence_name].setData(new_computational_sequences_data[sequence_name],sequence_name)
			if metadata_copy:
				#if there is no metadata for this computational sequences from the previous one or no previous computational sequenece
				if sequence_name not in list(old_metadata.keys()):
					log.error ("Metadata not available to copy ..., please provide metadata before writing to disk later", error =False)
				self.computational_sequences[sequence_name].setMetadata(old_metadata[sequence_name],sequence_name)
			self.computational_sequences[sequence_name].rootName=sequence_name

	def deploy(self,destination,filenames):
		if os.path.isdir(destination) is False:
			os.mkdir(destination)
		for seq_key in list(self.computational_sequences.keys()):
			if seq_key not in list(filenames.keys()):
				log.error("Filename for %s computational sequences not specified"%seq_key)
			filename=filenames[seq_key]
			if filename [:-4] != '.csd':
				filename+='.csd'
			self.computational_sequences[seq_key].deploy(os.path.join(destination,filename))

	def sort(self,sort_function=sorted):
		""" Sorts the computational sequence data based on the start time in intervals 

		#Arguments
			self: The dataset object.
			sort_function: The function passed for sorting. This function will receive the intervals one by one for each key 

		#Returns
			Does not return any values - replaces in place

		"""


		for key in list(self.keys()):
			for csd in list(self[key].keys()):
				this_intervals_np=numpy.array(self[key][csd]["intervals"])
				this_features_np=numpy.array(self[key][csd]["features"])
				sorted_indices=sort_function(range(this_intervals_np.shape[0]),key=lambda x: this_intervals_np[x,0])
				sorted_this_intervals_np=this_intervals_np[sorted_indices,:]
				sorted_this_features_np=this_features_np[sorted_indices,:]
				self[key][csd]["intervals"]=sorted_this_intervals_np
				self[key][csd]["features"]=sorted_this_features_np

	#TODO: Add folds to this function
	def get_tensors(self,seq_len,non_sequences=[],direction=False,folds=None):
		""" Returns trainable tensor from computational sequence data

		#Arguments
			self: The dataset object.
			seq_len: The maximum sequence length for the computational sequence entries, e.g. sentence length in words.
			direction: True for right padding and False for left padding.
			folds: The folds in which the dataset should be split.
		#Returns
			Dictionary of numpy arrays with the same data type as computational sequences. Dictionaries include the same keys as the dataset

		"""

		csds=list(self.keys())

		def handle_none_folds():
			return_folds={}
			for key in list(self[csds[0]].keys()):
				return_folds[key.split("[")[0]]=None
			return [list(return_folds.keys())]
	
		if folds==None:
			log.error("No fold specified for get_tensors, defaulting to the first computational sequence ids",error=False)
			folds=handle_none_folds()

		self.hard_unify()

		data=[]
		output=[]
	
		for i in range (len(folds)):
			data.append({})
			output.append({})
		
		def lpad(this_array,direction):
			if direction==False:
				temp_array=numpy.concatenate([numpy.zeros([seq_len]+list(this_array.shape[1:])),this_array],axis=0)[-seq_len:,...]
			else:
				temp_array=numpy.concatenate([this_array,numpy.zeros([seq_len]+list(this_array.shape[1:]))],axis=0)[:seq_len,...]
			return temp_array

		def detect_entry_fold(entry,folds):
			entry_id=entry.split("[")[0]
			for i in range(len(folds)):
				if entry_id in folds[i]: return i
			return None
			

		if len(csds)==0:
			log.error("Dataset is empty, cannot get tensors. Exiting ...!",error=True)

		for i in range (len(folds)):
			for csd in csds:
				data[i][csd]=[]

		for key in list(self[csds[0]].keys()):
			which_fold=detect_entry_fold(key,folds)
			if which_fold==None:
				log.error("Key %s doesn't belong to any fold ... "%str(key),error=False)
				continue
			for csd in list(self.keys()):
				this_array=self[csd][key]["features"]
				if csd in non_sequences:
					data[which_fold][csd].append(this_array)
				else:
					data[which_fold][csd].append(lpad(this_array,direction))

		for i in range(len(folds)):
			for csd in csds:
				output[i][csd]=numpy.array(data[i][csd])

		return output

	def __intersect_and_copy(self,ref,relevant_entry,epsilon):

		sub=relevant_entry["intervals"]
		features=relevant_entry["features"]

		#copying and inverting the ref
		ref_copy=ref.copy()
		ref_copy[1]=-ref_copy[1]
		ref_copy=ref_copy[::-1]
		sub_copy=sub.copy()
		sub_copy[:,0]=-sub_copy[:,0]
		#finding where intersect happens
		where_intersect=(numpy.all((sub_copy-ref_copy)>(-epsilon),axis=1)==True)
		intersectors=sub[where_intersect,:]
		intersectors=numpy.concatenate([numpy.maximum(intersectors[:,0],ref[0])[:,None],numpy.minimum(intersectors[:,1],ref[1])[:,None]],axis=1)
		intersectors_features=features[where_intersect,:]
		#checking for boundary cases and also zero length
		where_nonzero_len=numpy.where(abs(intersectors[:,0]-intersectors[:,1])>epsilon)
		intersectors_final=intersectors[where_nonzero_len]
		intersectors_features_final=intersectors_features[where_nonzero_len]
		return intersectors_final,intersectors_features_final

	#TODO: Need tqdm bar for this as well
	def __get_relevant_entries(self,reference):
		relevant_entries={}
		relevant_entries_np={}

		#pbar = tqdm(total=count,unit=" Computational Sequence Entries",leave=False)


		for otherseq_key in set(list(self.computational_sequences.keys()))-set([reference]):
			relevant_entries[otherseq_key]={}
			relevant_entries_np[otherseq_key]={}
			sub_compseq=self.computational_sequences[otherseq_key] 
			for key in list(sub_compseq.data.keys()):              
				keystripped=key.split('[')[0]                  
				if keystripped not in relevant_entries[otherseq_key]:                           
					relevant_entries[otherseq_key][keystripped]={}
					relevant_entries[otherseq_key][keystripped]["intervals"]=[]                     
					relevant_entries[otherseq_key][keystripped]["features"]=[]                                                            
		        
				relev_intervals=self.computational_sequences[otherseq_key].data[key]["intervals"]                                             
				relev_features=self.computational_sequences[otherseq_key].data[key]["features"]         
				if len(relev_intervals.shape)<2:
					relev_intervals=relev_intervals[None,:]
					relev_features=relev_features[None,:]

				relevant_entries[otherseq_key][keystripped]["intervals"].append(relev_intervals)
				relevant_entries[otherseq_key][keystripped]["features"].append(relev_features)
		                        
			for key in list(relevant_entries[otherseq_key].keys()):
				relev_intervals_np=numpy.concatenate(relevant_entries[otherseq_key][key]["intervals"],axis=0)                                 
				relev_features_np=numpy.concatenate(relevant_entries[otherseq_key][key]["features"],axis=0)
				sorted_indices=sorted(range(relev_intervals_np.shape[0]),key=lambda x: relev_intervals_np[x,0])                               
				relev_intervals_np=relev_intervals_np[sorted_indices,:]                         
				relev_features_np=relev_features_np[sorted_indices,:]

				relevant_entries_np[otherseq_key][key]={}
				relevant_entries_np[otherseq_key][key]["intervals"]=relev_intervals_np
				relevant_entries_np[otherseq_key][key]["features"]=relev_features_np
			log.status("Pre-alignment done for <%s> ..."%otherseq_key)
		return relevant_entries_np



import matplotlib.pyplot as plt
import h5py
import cv2
import tensorflow as tf
from keras.layers import Conv2D, Dense, LeakyReLU, Dropout, Input
from keras.layers import Reshape, Conv2DTranspose, Flatten
from keras.models import Model
from keras import optimizers
physical_devices = tf.config.list_physical_devices('GPU')
tf.config.experimental.set_memory_growth(physical_devices[0], True)
from keras.models import load_model
from numpy.random import randn
from matplotlib import pyplot
import numpy as np
import os
from tqdm import tqdm
from keras.preprocessing import image
from keras.preprocessing.image import ImageDataGenerator, array_to_img, img_to_array, load_img
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

class GanNode:

    def __init__(self,root_path,dim=32):
        self.d_loss=[]
        self.a_loss=[]
        self.root_path=root_path
        self.dim=dim
        if not os.path.exists(self.root_path):
            os.mkdir(self.root_path)

    def create_samples_from(self,src_img_path,sample_name="sample",num_samples=12000,use_gray=False):
        print("Creating samples from a picture...")
        datagen = ImageDataGenerator(rotation_range=40,
                                     width_shift_range=0.2,
                                     height_shift_range=0.2,
                                     shear_range=0.2,
                                     zoom_range=0.2,
                                     horizontal_flip=True,
                                     fill_mode='nearest')
        if use_gray:
            img = load_img(src_img_path,grayscale=True)
        else:
            img = load_img(src_img_path)
        # img = img.convert('L')
        x = img_to_array(img)
        x = x.reshape((1,) + x.shape)



        save_to_dir=self.root_path+"/data"

        if not os.path.exists(save_to_dir):
            os.mkdir(save_to_dir)

        i = 0
        for batch in datagen.flow(x, batch_size=200,
                                  save_to_dir=save_to_dir, save_prefix=sample_name, save_format='jpg'):
            i += 1
            if i > num_samples:
                break

    def get_files(self,file_dir):
        cats = []

        label_cats = []
        # dogs = []
        # label_dogs = []

        for file in os.listdir(file_dir ):
            cats.append(file_dir + '/' + file)
            label_cats.append(0)  # 添加标签，该类标签为0，此为2分类例子，多类别识别问题自行添加
        # for file in os.listdir(file_dir + '/1'):
        #    dogs.append(file_dir + '/1' + '/' + file)
        #    label_dogs.append(1)

        # 把cat和dog合起来组成一个list（img和lab）
        # image_list = np.hstack((cats, dogs))
        # label_list = np.hstack((label_cats, label_dogs))
        image_list = cats
        label_list = label_cats

        # 利用shuffle打乱顺序
        temp = np.array([image_list, label_list])
        temp = temp.transpose()
        np.random.shuffle(temp)

        # 从打乱的temp中再取出list（img和lab）
        image_list = list(temp[:, 0])
        label_list = list(temp[:, 1])
        label_list = [int(i) for i in label_list]

        return image_list, label_list
        # 返回两个list 分别为图片文件名及其标签 顺序已被打乱

    def create_dataset_gray(self,train_dir,out_path):

        # train_dir = '../data_preprocess/train'
        image_list, label_list = self.get_files(train_dir)

        # print(len(image_list))
        # print(len(label_list))

        # 450为数据长度的20%
        percent20 = int(len(image_list) * 0.2 )
        dim = self.dim
        Train_image = np.random.rand(len(image_list) - percent20, dim, dim).astype('float32')
        Train_label = np.random.rand(len(image_list) - percent20, 1).astype('float32')

        Test_image = np.random.rand(percent20, dim, dim).astype('float32')
        Test_label = np.random.rand(percent20, 1).astype('float32')

        print("train len: ",len(Train_image))
        print("test len: ", len(Test_image))
        for i in range(len(image_list) - percent20):
            img = plt.imread(image_list[i])
            # img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
            shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)

            Train_image[i] = np.array(shrink_img)
            Train_label[i] = np.array(label_list[i])

        for i in range(len(image_list) - percent20, len(image_list)):
            img = plt.imread(image_list[i])

            # img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
            shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)

            Test_image[i + percent20 - len(image_list)] = np.array(shrink_img)
            Test_label[i + percent20 - len(image_list)] = np.array(label_list[i])

        # Create a new file
        f = h5py.File(out_path, 'w')
        f.create_dataset('X_train', data=Train_image)
        f.create_dataset('y_train', data=Train_label)
        f.create_dataset('X_test', data=Test_image)
        f.create_dataset('y_test', data=Test_label)
        f.close()

    def create_dataset(self,train_dir,out_path,use_gray=False):

        # train_dir = '../data_preprocess/train'
        image_list, label_list = self.get_files(train_dir)

        # print(len(image_list))
        # print(len(label_list))

        # 450为数据长度的20%
        percent20 = int(len(image_list) * 0.2 )
        dim = self.dim
        Train_image = np.random.rand(len(image_list) - percent20, dim, dim, 3).astype('float32')
        Train_label = np.random.rand(len(image_list) - percent20, 1).astype('float32')

        Test_image = np.random.rand(percent20, dim, dim, 3).astype('float32')
        Test_label = np.random.rand(percent20, 1).astype('float32')

        print("train len: ",len(Train_image))
        print("test len: ", len(Test_image))
        for i in range(len(image_list) - percent20):
            img = plt.imread(image_list[i])
            if use_gray:
                img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
                shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)
            else:
                shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)
            Train_image[i] = np.array(shrink_img)
            Train_label[i] = np.array(label_list[i])

        for i in range(len(image_list) - percent20, len(image_list)):
            img = plt.imread(image_list[i])
            if use_gray:
                img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
                shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)
            else:
                shrink_img = cv2.resize(img, (dim, dim), interpolation=cv2.INTER_AREA)
            Test_image[i + percent20 - len(image_list)] = np.array(shrink_img)
            Test_label[i + percent20 - len(image_list)] = np.array(label_list[i])

        # Create a new file
        f = h5py.File(out_path, 'w')
        f.create_dataset('X_train', data=Train_image)
        f.create_dataset('y_train', data=Train_label)
        f.create_dataset('X_test', data=Test_image)
        f.create_dataset('y_test', data=Test_label)
        f.close()

    def build64(self,data_path,model_path,save_dir,iterations=1000,batch_size=64,height=32,width=32,latent_dim=100,channels=3):

        """设置相关参数"""
        # 潜变量维度
        # latent_dim = 100
        # 输入像素维度
        # height = 64
        #width = 64
        # channels = 3

        # 搭建生成器网络（Model方式）
        generator_input = Input(shape=(latent_dim,))
        print("shape1", generator_input.shape)
        x = Dense(128 * 32 * 32)(generator_input)
        print("shape2", x.shape)
        x = LeakyReLU()(x)
        print("shape3", x.shape)
        x = Reshape((32, 32, 128))(x)
        print("shape4", x.shape)

        # IN: 16*16*128  OUT: 16*16*256
        x = Conv2D(256, 5, padding='same')(x)
        print("shape5", x.shape)
        x = LeakyReLU()(x)
        print("shape6", x.shape)

        # IN: 16*16*256  OUT: 32*32*256
        x = Conv2DTranspose(256, 4, strides=2, padding='same')(x)
        x = LeakyReLU()(x)
        print("shape7", x.shape)

        # 通过反卷积操作，把维度变得和图片一样大了，就别用反卷积了
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)
        print("shape8", x.shape)

        # 把通道变回来
        x = Conv2D(channels, 7, activation='tanh', padding='same')(x)
        print("shape9", x.shape)
        generator = Model(generator_input, x)
        # generator.summary()

        # 搭建判别器，判别器就真的是一个卷积神经网络了
        discriminator_input = Input(shape=(height, width, channels))
        # IN:32*32*3   OUT: 30*30*128
        x = Conv2D(128, 3)(discriminator_input)
        x = LeakyReLU()(x)

        # IN: 30*30*128  OUT:14*14*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:14*14*128   OUT:6*6*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:6*6*128  OUT:2*2*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # 展开成512个神经元
        x = Flatten()(x)
        x = Dropout(0.4)(x)
        x = Dense(1, activation='sigmoid')(x)

        discriminator = Model(discriminator_input, x)
        # discriminator.summary()

        discriminator_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0008,
                                                     clipvalue=1.0,
                                                     decay=1e-8)

        discriminator.compile(optimizer=discriminator_optimizer,
                              loss='binary_crossentropy')

        # 将上面两个连起来组成生成对抗网络
        # 将判别器参数设置为不可训练
        discriminator.trainable = False

        gan_input = Input(shape=(latent_dim,))
        gan_output = discriminator(generator(gan_input))
        # 搭建对抗网络
        gan = Model(gan_input, gan_output)
        gan_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0004,
                                           clipvalue=1.0,
                                           decay=1e-8)
        gan.compile(optimizer=gan_optimizer, loss='binary_crossentropy')

        ## 下面导入数据集进行训练，数据集选用的是Keras自带的CIFAR-10数据集
        import os
        from keras.preprocessing import image

        # 加载数据集（手写数字的数据集也可以这样加载）
        # (x_train, y_train), (_, _) = keras.datasets.cifar10.load_data()
        # 指定青蛙图像（编号为6）
        # x_train = x_train[y_train.flatten() == 6]
        # Load hdf5 dataset

        import h5py
        train_dataset = h5py.File(data_path, 'r')
        train_set_x_orig = np.array(train_dataset['X_train'][:])  # your train set features
        train_set_y_orig = np.array(train_dataset['y_train'][:])  # your train set labels
        test_set_x_orig = np.array(train_dataset['X_test'][:])  # your train set features
        test_set_y_orig = np.array(train_dataset['y_test'][:])  # your train set labels
        train_dataset.close()

        print(train_set_x_orig.shape)
        print(train_set_y_orig.shape)

        print(train_set_x_orig.max())
        print(train_set_x_orig.min())

        print(test_set_x_orig.shape)
        print(test_set_y_orig.shape)
        x_train = train_set_x_orig

        print("data size:", x_train.shape)
        x_train = x_train.reshape((x_train.shape[0],) + (height, width, channels)).astype('float32') / 255.

        print("data size(adjusted):", x_train.shape)

        # 下面开始训练
        # iterations = 10000
        # batch_size = 32
        #save_dir = 'image64'

        start = 0
        for step in range(iterations):

            # 先通过噪声生成64张伪造图片
            noise = np.random.normal(size=(batch_size, latent_dim))
            images_fake = generator.predict(noise)

            # 从真实图片中抽64张
            end = start + batch_size
            images_train = x_train[start:end]

            # 两者混合作为训练集，并打上标签
            x = np.concatenate([images_fake, images_train])
            y = np.concatenate([np.zeros((batch_size, 1)), np.ones((batch_size, 1))])

            # 向标签中添加噪声
            y += 0.05 * np.random.random(y.shape)

            # 训练判别器
            d_loss = discriminator.train_on_batch(x, y)

            # 训练生成对抗网络
            noise = np.random.normal(size=(batch_size, latent_dim))
            labels = np.ones((batch_size, 1))

            # 通过gan模型来训练生成器模型，冻结判别器模型权重
            a_loss = gan.train_on_batch(noise, labels)

            start += batch_size
            if start > len(x_train) - batch_size:
                start = 0

            # 每500步绘图并保存
            if step % 100 == 0:
                generator.save(model_path)
            # if step % 10 == 0:
            print(f"Iteration={step + 1}\tDLoss={d_loss}\tALoss={a_loss}")
            # print('discriminator loss:', d_loss)
            # print('adversarial loss:', a_loss)
            self.d_loss.append(d_loss)
            self.a_loss.append(a_loss)
            img = image.array_to_img(images_fake[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'generated_sample' + str(step) + '.png'))
            img = image.array_to_img(images_train[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'real_sample' + str(step) + '.png'))

    def build128(self,data_path,model_path,save_dir,iterations=1000,batch_size=64,height=32,width=32,latent_dim=100,channels=3):

        """设置相关参数"""
        # 潜变量维度
        # latent_dim = 100
        # 输入像素维度
        # height = 128
        # width = 128
        # channels = 3

        # 搭建生成器网络（Model方式）
        generator_input = Input(shape=(latent_dim,))
        x = Dense(128 * 64 * 64)(generator_input)
        x = LeakyReLU()(x)
        x = Reshape((64, 64, 128))(x)

        # IN: 16*16*128  OUT: 16*16*256
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)

        # IN: 16*16*256  OUT: 32*32*256
        x = Conv2DTranspose(256, 4, strides=2, padding='same')(x)
        x = LeakyReLU()(x)

        # 通过反卷积操作，把维度变得和图片一样大了，就别用反卷积了
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)

        # 把通道变回来
        x = Conv2D(channels, 7, activation='tanh', padding='same')(x)

        generator = Model(generator_input, x)
        # generator.summary()

        # 搭建判别器，判别器就真的是一个卷积神经网络了
        discriminator_input = Input(shape=(height, width, channels))
        # IN:32*32*3   OUT: 30*30*128
        x = Conv2D(128, 3)(discriminator_input)
        x = LeakyReLU()(x)

        # IN: 30*30*128  OUT:14*14*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:14*14*128   OUT:6*6*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:6*6*128  OUT:2*2*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # 展开成512个神经元
        x = Flatten()(x)
        x = Dropout(0.4)(x)
        x = Dense(1, activation='sigmoid')(x)

        discriminator = Model(discriminator_input, x)
        # discriminator.summary()

        discriminator_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0008,
                                                     clipvalue=1.0,
                                                     decay=1e-8)

        discriminator.compile(optimizer=discriminator_optimizer,
                              loss='binary_crossentropy')

        # 将上面两个连起来组成生成对抗网络
        # 将判别器参数设置为不可训练
        discriminator.trainable = False

        gan_input = Input(shape=(latent_dim,))
        gan_output = discriminator(generator(gan_input))
        # 搭建对抗网络
        gan = Model(gan_input, gan_output)
        gan_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0004,
                                           clipvalue=1.0,
                                           decay=1e-8)
        gan.compile(optimizer=gan_optimizer, loss='binary_crossentropy')

        ## 下面导入数据集进行训练，数据集选用的是Keras自带的CIFAR-10数据集
        import os
        from keras.preprocessing import image

        # 加载数据集（手写数字的数据集也可以这样加载）
        # (x_train, y_train), (_, _) = keras.datasets.cifar10.load_data()
        # 指定青蛙图像（编号为6）
        # x_train = x_train[y_train.flatten() == 6]
        # Load hdf5 dataset

        import h5py
        train_dataset = h5py.File(data_path, 'r')
        train_set_x_orig = np.array(train_dataset['X_train'][:])  # your train set features
        train_set_y_orig = np.array(train_dataset['y_train'][:])  # your train set labels
        test_set_x_orig = np.array(train_dataset['X_test'][:])  # your train set features
        test_set_y_orig = np.array(train_dataset['y_test'][:])  # your train set labels
        train_dataset.close()

        print(train_set_x_orig.shape)
        print(train_set_y_orig.shape)

        print(train_set_x_orig.max())
        print(train_set_x_orig.min())

        print(test_set_x_orig.shape)
        print(test_set_y_orig.shape)
        x_train = train_set_x_orig

        print("data size:", x_train.shape)
        x_train = x_train.reshape((x_train.shape[0],) + (height, width, channels)).astype('float32') / 255.

        print("data size(adjusted):", x_train.shape)

        # 下面开始训练
        # iterations = 10000
        # batch_size = 16
        # save_dir = 'image128'

        start = 0
        for step in range(iterations):

            # 先通过噪声生成64张伪造图片
            noise = np.random.normal(size=(batch_size, latent_dim))
            images_fake = generator.predict(noise)

            # 从真实图片中抽64张
            end = start + batch_size
            images_train = x_train[start:end]

            # 两者混合作为训练集，并打上标签
            x = np.concatenate([images_fake, images_train])
            y = np.concatenate([np.zeros((batch_size, 1)), np.ones((batch_size, 1))])

            # 向标签中添加噪声
            y += 0.05 * np.random.random(y.shape)

            # 训练判别器
            d_loss = discriminator.train_on_batch(x, y)

            # 训练生成对抗网络
            noise = np.random.normal(size=(batch_size, latent_dim))
            labels = np.ones((batch_size, 1))

            # 通过gan模型来训练生成器模型，冻结判别器模型权重
            a_loss = gan.train_on_batch(noise, labels)

            start += batch_size
            if start > len(x_train) - batch_size:
                start = 0

            # 每500步绘图并保存
            if step % 100 == 0:
                generator.save(model_path)
            # if step % 10 == 0:
            print(f"Iteration={step + 1}\tDLoss={d_loss}\tALoss={a_loss}")
            # print('discriminator loss:', d_loss)
            # print('adversarial loss:', a_loss)
            self.d_loss.append(d_loss)
            self.a_loss.append(a_loss)
            img = image.array_to_img(images_fake[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'generated_sample' + str(step) + '.png'))
            img = image.array_to_img(images_train[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'real_sample' + str(step) + '.png'))

    def build_gray(self,data_path,model_path,save_dir,iterations=1000,batch_size=64,height=32,width=32,latent_dim=100,channels=1):
        print("Building model...")
        """设置相关参数"""
        # 潜变量维度
        # latent_dim = 100
        # 输入像素维度
        # height = 32
        # width = 32
        # channels = 3

        # 搭建生成器网络（Model方式）
        generator_input = Input(shape=(latent_dim,))
        x = Dense(128 * 16 * 16)(generator_input)
        x = LeakyReLU()(x)
        x = Reshape((16, 16, 128))(x)

        # IN: 16*16*128  OUT: 16*16*256
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)

        # IN: 16*16*256  OUT: 32*32*256
        x = Conv2DTranspose(256, 4, strides=2, padding='same')(x)
        x = LeakyReLU()(x)

        # 通过反卷积操作，把维度变得和图片一样大了，就别用反卷积了
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)

        # 把通道变回来
        x = Conv2D(channels, 7, activation='tanh', padding='same')(x)

        generator = Model(generator_input, x)
        # generator.summary()

        # 搭建判别器，判别器就真的是一个卷积神经网络了
        discriminator_input = Input(shape=(height, width, channels))
        # IN:32*32*3   OUT: 30*30*128
        x = Conv2D(128, channels)(discriminator_input)
        x = LeakyReLU()(x)

        # IN: 30*30*128  OUT:14*14*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:14*14*128   OUT:6*6*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:6*6*128  OUT:2*2*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # 展开成512个神经元
        x = Flatten()(x)
        x = Dropout(0.4)(x)
        x = Dense(1, activation='sigmoid')(x)

        discriminator = Model(discriminator_input, x)
        # discriminator.summary()

        discriminator_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0008,
                                                                clipvalue=1.0,
                                                                decay=1e-8)

        discriminator.compile(optimizer=discriminator_optimizer,
                              loss='binary_crossentropy')

        # 将上面两个连起来组成生成对抗网络
        # 将判别器参数设置为不可训练
        discriminator.trainable = False

        gan_input = Input(shape=(latent_dim,))
        gan_output = discriminator(generator(gan_input))
        # 搭建对抗网络
        gan = Model(gan_input, gan_output)
        gan_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0004,
                                                      clipvalue=1.0,
                                                      decay=1e-8)
        gan.compile(optimizer=gan_optimizer, loss='binary_crossentropy')

        ## 下面导入数据集进行训练，数据集选用的是Keras自带的CIFAR-10数据集


        # 加载数据集（手写数字的数据集也可以这样加载）
        # (x_train, y_train), (_, _) = keras.datasets.cifar10.load_data()
        # 指定青蛙图像（编号为6）
        # x_train = x_train[y_train.flatten() == 6]
        # Load hdf5 dataset

        train_dataset = h5py.File(data_path, 'r')
        train_set_x_orig = np.array(train_dataset['X_train'][:])  # your train set features
        train_set_y_orig = np.array(train_dataset['y_train'][:])  # your train set labels
        test_set_x_orig = np.array(train_dataset['X_test'][:])  # your train set features
        test_set_y_orig = np.array(train_dataset['y_test'][:])  # your train set labels
        train_dataset.close()

        print(train_set_x_orig.shape)
        print(train_set_y_orig.shape)

        print(train_set_x_orig.max())
        print(train_set_x_orig.min())

        print(test_set_x_orig.shape)
        print(test_set_y_orig.shape)
        x_train = train_set_x_orig

        print("data size:", x_train.shape)
        x_train = x_train.reshape((x_train.shape[0],) + (height, width, channels)).astype('float32') / 255.

        print("data size(adjusted):", x_train.shape)

        # 下面开始训练
        # iterations = 10000
        # batch_size = 64
        # save_dir = 'outputs/images32'


        start = 0
        for step in range(iterations):

            # 先通过噪声生成64张伪造图片
            noise = np.random.normal(size=(batch_size, latent_dim))
            images_fake = generator.predict(noise)

            # 从真实图片中抽64张
            end = start + batch_size
            images_train = x_train[start:end]

            # 两者混合作为训练集，并打上标签
            x = np.concatenate([images_fake, images_train])
            y = np.concatenate([np.zeros((batch_size, 1)), np.ones((batch_size, 1))])

            # 向标签中添加噪声
            y += 0.05 * np.random.random(y.shape)

            # 训练判别器
            d_loss = discriminator.train_on_batch(x, y)

            # 训练生成对抗网络
            noise = np.random.normal(size=(batch_size, latent_dim))
            labels = np.ones((batch_size, 1))

            # 通过gan模型来训练生成器模型，冻结判别器模型权重
            a_loss = gan.train_on_batch(noise, labels)

            start += batch_size
            if start > len(x_train) - batch_size:
                start = 0

            # 每500步绘图并保存
            if step % 100 == 0:
                generator.save(model_path)
            # if step % 10 == 0:
            print(f"Iteration={step+1}\tDLoss={d_loss}\tALoss={a_loss}")
            # print('discriminator loss:', d_loss)
            # print('adversarial loss:', a_loss)
            self.d_loss.append(d_loss)
            self.a_loss.append(a_loss)
            img = image.array_to_img(images_fake[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'generated_sample' + str(step) + '.png'))
            img = image.array_to_img(images_train[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'real_sample' + str(step) + '.png'))


    def build(self,data_path,model_path,save_dir,iterations=1000,batch_size=64,height=32,width=32,latent_dim=100,channels=3):
        print("Building model...")
        """设置相关参数"""
        # 潜变量维度
        # latent_dim = 100
        # 输入像素维度
        # height = 32
        # width = 32
        # channels = 3

        # 搭建生成器网络（Model方式）
        generator_input = Input(shape=(latent_dim,))
        x = Dense(128 * 16 * 16)(generator_input)
        x = LeakyReLU()(x)
        x = Reshape((16, 16, 128))(x)

        # IN: 16*16*128  OUT: 16*16*256
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)

        # IN: 16*16*256  OUT: 32*32*256
        x = Conv2DTranspose(256, 4, strides=2, padding='same')(x)
        x = LeakyReLU()(x)

        # 通过反卷积操作，把维度变得和图片一样大了，就别用反卷积了
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)
        x = Conv2D(256, 5, padding='same')(x)
        x = LeakyReLU()(x)

        # 把通道变回来
        x = Conv2D(channels, 7, activation='tanh', padding='same')(x)

        generator = Model(generator_input, x)
        # generator.summary()

        # 搭建判别器，判别器就真的是一个卷积神经网络了
        discriminator_input = Input(shape=(height, width, channels))
        # IN:32*32*3   OUT: 30*30*128
        x = Conv2D(128, 3)(discriminator_input)
        x = LeakyReLU()(x)

        # IN: 30*30*128  OUT:14*14*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:14*14*128   OUT:6*6*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # IN:6*6*128  OUT:2*2*128
        x = Conv2D(128, 4, strides=2)(x)
        x = LeakyReLU()(x)

        # 展开成512个神经元
        x = Flatten()(x)
        x = Dropout(0.4)(x)
        x = Dense(1, activation='sigmoid')(x)

        discriminator = Model(discriminator_input, x)
        # discriminator.summary()

        discriminator_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0008,
                                                                clipvalue=1.0,
                                                                decay=1e-8)

        discriminator.compile(optimizer=discriminator_optimizer,
                              loss='binary_crossentropy')

        # 将上面两个连起来组成生成对抗网络
        # 将判别器参数设置为不可训练
        discriminator.trainable = False

        gan_input = Input(shape=(latent_dim,))
        gan_output = discriminator(generator(gan_input))
        # 搭建对抗网络
        gan = Model(gan_input, gan_output)
        gan_optimizer = optimizers.rmsprop_v2.RMSprop(lr=0.0004,
                                                      clipvalue=1.0,
                                                      decay=1e-8)
        gan.compile(optimizer=gan_optimizer, loss='binary_crossentropy')

        ## 下面导入数据集进行训练，数据集选用的是Keras自带的CIFAR-10数据集


        # 加载数据集（手写数字的数据集也可以这样加载）
        # (x_train, y_train), (_, _) = keras.datasets.cifar10.load_data()
        # 指定青蛙图像（编号为6）
        # x_train = x_train[y_train.flatten() == 6]
        # Load hdf5 dataset

        train_dataset = h5py.File(data_path, 'r')
        train_set_x_orig = np.array(train_dataset['X_train'][:])  # your train set features
        train_set_y_orig = np.array(train_dataset['y_train'][:])  # your train set labels
        test_set_x_orig = np.array(train_dataset['X_test'][:])  # your train set features
        test_set_y_orig = np.array(train_dataset['y_test'][:])  # your train set labels
        train_dataset.close()

        print(train_set_x_orig.shape)
        print(train_set_y_orig.shape)

        print(train_set_x_orig.max())
        print(train_set_x_orig.min())

        print(test_set_x_orig.shape)
        print(test_set_y_orig.shape)
        x_train = train_set_x_orig

        print("data size:", x_train.shape)
        x_train = x_train.reshape((x_train.shape[0],) + (height, width, channels)).astype('float32') / 255.

        print("data size(adjusted):", x_train.shape)

        # 下面开始训练
        # iterations = 10000
        # batch_size = 64
        # save_dir = 'outputs/images32'


        start = 0
        for step in range(iterations):

            # 先通过噪声生成64张伪造图片
            noise = np.random.normal(size=(batch_size, latent_dim))
            images_fake = generator.predict(noise)

            # 从真实图片中抽64张
            end = start + batch_size
            images_train = x_train[start:end]

            # 两者混合作为训练集，并打上标签
            x = np.concatenate([images_fake, images_train])
            y = np.concatenate([np.zeros((batch_size, 1)), np.ones((batch_size, 1))])

            # 向标签中添加噪声
            y += 0.05 * np.random.random(y.shape)

            # 训练判别器
            d_loss = discriminator.train_on_batch(x, y)

            # 训练生成对抗网络
            noise = np.random.normal(size=(batch_size, latent_dim))
            labels = np.ones((batch_size, 1))

            # 通过gan模型来训练生成器模型，冻结判别器模型权重
            a_loss = gan.train_on_batch(noise, labels)

            start += batch_size
            if start > len(x_train) - batch_size:
                start = 0

            # 每500步绘图并保存
            if step % 100 == 0:
                generator.save(model_path)
            # if step % 10 == 0:
            print(f"Iteration={step+1}\tDLoss={d_loss}\tALoss={a_loss}")
            # print('discriminator loss:', d_loss)
            # print('adversarial loss:', a_loss)
            self.d_loss.append(d_loss)
            self.a_loss.append(a_loss)
            img = image.array_to_img(images_fake[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'generated_sample' + str(step) + '.png'))
            img = image.array_to_img(images_train[0] * 255., scale=False)
            img.save(os.path.join(save_dir, 'real_sample' + str(step) + '.png'))

    # example of loading the generator model and generating images

    # generate points in latent space as input for the generator
    def generate_latent_points(self,latent_dim, n_samples):
        # generate points in the latent space
        x_input = randn(latent_dim * n_samples)
        # reshape into a batch of inputs for the network
        x_input = x_input.reshape(n_samples, latent_dim)
        return x_input

    # create and save a plot of generated images (reversed grayscale)
    def show_plot(self,examples, n):
        # plot images
        for i in range(n * n):
            # define subplot
            pyplot.subplot(n, n, 1 + i)
            # turn off axis
            pyplot.axis('off')
            # plot raw pixel data
            # pyplot.imshow((examples[i] * 255).astype(np.uint8), cmap='jet')
            pyplot.imshow((examples[i] * 255).astype(np.uint8),cmap='gray')
        pyplot.show()


    def generate(self,n_samples=10,latent_dim=100,show=True):
        print("Generating sample from model")
        model_path=self.root_path+"/model"
        # load model
        model = load_model(model_path)
        model.summary()
        # generate images
        latent_points = self.generate_latent_points(latent_dim, n_samples*n_samples)
        # noise = np.random.normal(size=(64, 100))
        # generate images
        X = model.predict(latent_points)
        # plot the result
        if show:
            self.show_plot(X, n_samples)
        return  X

    def build_model(self,iterations=1000,train_dir="",use_gray=False):
        if train_dir=="":
            train_dir = f"{self.root_path}/data"
        out_path = f"{self.root_path}/data.h5"
        model_path = f"{self.root_path}/model"
        generated_path = f"{self.root_path}/generated"
        if not os.path.exists(model_path):
            os.mkdir(model_path)
        if not os.path.exists(generated_path):
            os.mkdir(generated_path)

        if not use_gray:
            self.create_dataset(train_dir=train_dir, out_path=out_path,use_gray=False)
        else:
            self.create_dataset_gray(train_dir=train_dir, out_path=out_path)

        if self.dim==32:
            if use_gray:
                self.build_gray(data_path=out_path, model_path=model_path, save_dir=generated_path, iterations=iterations,
                           width=self.dim, height=self.dim,channels=1)
            else:
                self.build(data_path=out_path, model_path=model_path, save_dir=generated_path,iterations=iterations,width=self.dim,height=self.dim)
        elif self.dim==64:
            if use_gray:
                self.build64(data_path=out_path, model_path=model_path, save_dir=generated_path, iterations=iterations,
                           width=self.dim, height=self.dim,channels=1)
            else:
                self.build64(data_path=out_path, model_path=model_path, save_dir=generated_path, iterations=iterations,
                       width=self.dim, height=self.dim)
        elif self.dim==128:
            self.build128(data_path=out_path, model_path=model_path, save_dir=generated_path, iterations=iterations,
                         width=self.dim, height=self.dim)
        else:
            raise Exception("not found dim!")
        # self.generate(model_path=model_path)

    def export_performance(self,export_file=""):
        if export_file=="":
            export_folder=self.root_path+"/performance"
            if not os.path.exists(export_folder):
                os.mkdir(export_folder)
            export_file=f"{export_folder}/loss.csv"
        f_out=open(export_file,"w",encoding='utf-8')
        f_out.write("Iteration\tLoss(A)\tLoss(D)\n")
        for idx,a in enumerate(self.a_loss):
            f_out.write(f"{idx}\t{self.a_loss[idx]}\t{self.d_loss[idx]}\n")
        f_out.close()










import numpy as np
from scipy import misc
import random
import skimage
import skimage.io
import skimage.transform
import cv2
from matplotlib.image import imread,imsave
from PIL import Image

class ImageUtils:
	def __init__(self):
		pass

	def load_image_array(self,image_file, image_size):
		img = skimage.io.imread(image_file)
		# GRAYSCALE
		if len(img.shape) == 2:
			img_new = np.ndarray((img.shape[0], img.shape[1], 3), dtype='uint8')
			img_new[:, :, 0] = img
			img_new[:, :, 1] = img
			img_new[:, :, 2] = img
			img = img_new

		img_resized = skimage.transform.resize(img, (image_size, image_size))

		# FLIP HORIZONTAL WIRH A PROBABILITY 0.5
		if random.random() > 0.5:
			img_resized = np.fliplr(img_resized)

		return img_resized.astype('float32')

	def get_numpy_array_rgb(self,image_file):
		img = imread(image_file)
		return img

	def save_numpy_array_rgb(self,image_file,img_data):
		imsave(image_file,img_data)

	def resize_image(self,image_file,new_size,output_file=None,format="JPEG"):
		im = Image.open(image_file)
		im.thumbnail(new_size, Image.ANTIALIAS)
		if output_file!=None:
			im.save(output_file, format)
		return np.array(im)





import os
from typing import Dict, Tuple, Union

import mmkfeatures.video.models.r21d.transforms.rgb_transforms as T
import numpy as np
import torch
from torchvision.io.video import read_video
# import traceback
from torchvision.models.video import r2plus1d_18
from torchvision.transforms import Compose
from tqdm import tqdm
from mmkfeatures.video.utils.utils import (action_on_extraction, form_list_from_user_input,
                         form_slices, reencode_video_with_diff_fps, show_predictions_on_dataset)

PRE_CENTRAL_CROP_SIZE = (128, 171)
KINETICS_MEAN = [0.43216, 0.394666, 0.37645]
KINETICS_STD = [0.22803, 0.22145, 0.216989]
CENTRAL_CROP_MIN_SIDE_SIZE = 112
DEFAULT_R21D_STEP_SIZE = 16
DEFAULT_R21D_STACK_SIZE = 16

class ExtractR21D(torch.nn.Module):

    def __init__(self, args):
        super(ExtractR21D, self).__init__()
        self.feature_type = args.feature_type
        self.path_list = form_list_from_user_input(args)
        self.central_crop_min_side_size = CENTRAL_CROP_MIN_SIDE_SIZE
        self.extraction_fps = args.extraction_fps
        self.step_size = args.step_size
        self.stack_size = args.stack_size
        if self.step_size is None:
            self.step_size = DEFAULT_R21D_STEP_SIZE
        if self.stack_size is None:
            self.stack_size = DEFAULT_R21D_STACK_SIZE
        self.transforms = Compose([
            T.ToFloatTensorInZeroOne(),
            T.Resize(PRE_CENTRAL_CROP_SIZE),
            T.Normalize(mean=KINETICS_MEAN, std=KINETICS_STD),
            T.CenterCrop((CENTRAL_CROP_MIN_SIDE_SIZE, CENTRAL_CROP_MIN_SIDE_SIZE))
        ])
        self.show_pred = args.show_pred
        self.keep_tmp_files = args.keep_tmp_files
        self.on_extraction = args.on_extraction
        # not used, create an issue if you would like to save the frames
        self.tmp_path = os.path.join(args.tmp_path, self.feature_type)
        self.output_path = os.path.join(args.output_path, self.feature_type)
        self.progress = tqdm(total=len(self.path_list))

    def forward(self, indices: torch.LongTensor):
        '''
        Arguments:
            indices {torch.LongTensor} -- indices to self.path_list
        '''
        device = indices.device
        model, class_head = self.load_model(device)

        for idx in indices:
            # when error occurs might fail silently when run from torch data parallel
            try:
                feats_dict = self.extract(device, model, class_head, self.path_list[idx])
                action_on_extraction(feats_dict, self.path_list[idx], self.output_path, self.on_extraction)
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except Exception as e:
                # prints only the last line of an error. Use `traceback.print_exc()` for the whole traceback
                # traceback.print_exc()
                print(e)
                print(f'Extraction failed at: {self.path_list[idx]} with error (↑). Continuing extraction')

            # update tqdm progress bar
            self.progress.update()

    def extract(self, device: torch.device, model: torch.nn.Module, classifier: torch.nn.Module,
                video_path: Union[str, None] = None
                ) -> Dict[str, np.ndarray]:
        '''The extraction call. Made to clean the forward call a bit.

        Arguments:
            device {torch.device}
            model {torch.nn.Module}
            classifier {torch.nn.Module} -- pre-trained classification layer, will be used if
                                            show_pred is True

        Keyword Arguments:
            video_path {Union[str, None]} -- if you would like to use import it and use it as
                                             "path -> model features"-fashion (default: {None})

        Returns:
            Dict[str, np.ndarray] -- the dict with numpy feature
        '''
        # take the video, change fps and save to the tmp folder
        if self.extraction_fps is not None:
            video_path = reencode_video_with_diff_fps(video_path, self.tmp_path, self.extraction_fps)

        # read a video
        rgb, audio, info = read_video(video_path, pts_unit='sec')
        # prepare data (first -- transform, then -- unsqueeze)
        rgb = self.transforms(rgb)
        rgb = rgb.unsqueeze(0)
        # slice the stack of frames
        slices = form_slices(rgb.size(2), self.stack_size, self.step_size)

        vid_feats = []

        for stack_idx, (start_idx, end_idx) in enumerate(slices):
            # inference
            with torch.no_grad():
                output = model(rgb[:, :, start_idx:end_idx, :, :].to(device))
                vid_feats.extend(output.tolist())

                # show predicitons on kinetics dataset (might be useful for debugging)
                if self.show_pred:
                    logits = classifier(output)
                    print(f'{video_path} @ frames ({start_idx}, {end_idx})')
                    show_predictions_on_dataset(logits, 'kinetics')

        feats_dict = {
            self.feature_type: np.array(vid_feats),
        }

        return feats_dict

    def load_model(self, device: torch.device) -> Tuple[torch.nn.Module]:
        '''Defines the models, loads checkpoints, sends them to the device.

        Args:
            device (torch.device): The device

        Raises:
            NotImplementedError: if flow type is not implemented.

        Returns:
            Tuple[torch.nn.Module]: the model with identity head, the original classifier
        '''
        model = r2plus1d_18(pretrained=True)
        model = model.to(device)
        model.eval()
        # save the pre-trained classifier for show_preds and replace it in the net with identity
        class_head = model.fc
        model.fc = torch.nn.Identity()
        return model, class_head



import torch
from torch.utils.data import Dataset
from mmkfeatures.models.image_text_fusion.char_embedding.char_cnn_rnn import prepare_text


class MultimodalDataset(Dataset):
    # Arguments:
    # dataset_json (json): json contains dataset item paths.
    # data_folder (String): dataset root path.
    def __init__(self, dataset_json, data_folder, device):
        super().__init__()
        self.dataset_json = dataset_json
        self.data = []
        self.device = device
        for item in self.dataset_json:
            # Load saved image tensor and convert
            # text to encoded tensor, size 1024
            image = torch.load(data_folder + item['img'])
            image.requires_grad = False
            text = prepare_text(item['text'])
            image = image.to(self.device)
            text = text.to(self.device)
            self.data.append({'img': image, 'text': text})

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]
